from ._build_distance_matrix_biopython import build_distance_matrix_biopython
from ._build_distance_matrix_numpy import build_distance_matrix_numpy
from ._calc_rank_of_earliest_detectable_mrca_among import (
    calc_rank_of_earliest_detectable_mrca_among,
)
from ._calc_rank_of_mrca_bounds_among import calc_rank_of_mrca_bounds_among
from ._calc_rank_of_mrca_uncertainty_among import (
    calc_rank_of_mrca_uncertainty_among,
)
from ._does_definitively_share_no_common_ancestor import (
    does_definitively_share_no_common_ancestor,
)
from ._does_share_any_common_ancestor import does_share_any_common_ancestor

__all__ = [
    "build_distance_matrix_biopython",
    "build_distance_matrix_numpy",
    "calc_rank_of_earliest_detectable_mrca_among",
    "calc_rank_of_mrca_bounds_among",
    "calc_rank_of_mrca_uncertainty_among",
    "does_definitively_share_no_common_ancestor",
    "does_share_any_common_ancestor",
]



import unittest

from hstrat._auxiliary_lib import is_strictly_increasing


class TestIsStrictlyIncreasing(unittest.TestCase):

    # tests can run independently
    _multiprocess_can_split_ = True

    def test_empty(self):
        assert is_strictly_increasing([])

    def test_singleton(self):
        assert is_strictly_increasing(["a"])
        assert is_strictly_increasing([0])
        assert is_strictly_increasing([1])

    def test_nonincreasing(self):
        assert is_strictly_increasing(
            [
                *range(10),
            ]
        )
        assert not is_strictly_increasing(
            [
                0,
                *range(10),
            ]
        )
        assert not is_strictly_increasing(
            [
                0,
                0,
                *range(10),
                *range(9, 18),
            ]
        )

    def test_decreasing(self):
        assert not is_strictly_increasing(
            [
                0,
                -1,
            ]
        )
        assert not is_strictly_increasing(
            [
                *range(10),
                *range(2),
            ]
        )

    def test_increasing(self):
        assert is_strictly_increasing(
            [
                -1,
                0,
            ]
        )
        assert is_strictly_increasing(range(10))
        assert is_strictly_increasing([(-1, 0.0), (0, -2), (0, -1)])


if __name__ == "__main__":
    unittest.main()



import itertools as it
import random

from iterify import cyclify, iterify
import opytional as opyt
import pytest
from scipy import stats

from hstrat import hstrat


@pytest.mark.filterwarnings(
    "ignore:Insufficient common ranks between columns to detect common ancestry at given confidence level."
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        pytest.param(
            hstrat.HereditaryStratumOrderedStoreDict,
            marks=pytest.mark.heavy,
        ),
        hstrat.HereditaryStratumOrderedStoreList,
    ],
)
@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_comparison_commutativity_asynchronous(
    retention_policy,
    ordered_store,
    wrap,
):
    population = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_ordered_store=ordered_store,
            stratum_retention_policy=retention_policy,
        )
        for __ in range(10)
    ]

    for _generation in range(100):
        _ = _generation
        for first, second in it.combinations(population, 2):
            # assert commutativity
            assert (
                hstrat.calc_rank_of_mrca_bounds_among(
                    [wrap(first), wrap(second)], prior="arbitrary"
                )
                == hstrat.calc_rank_of_mrca_bounds_among(
                    [wrap(second), wrap(first)], prior="arbitrary"
                )
                == hstrat.calc_rank_of_mrca_bounds_among(
                    [wrap(second), wrap(first)] * 3, prior="arbitrary"
                )
            )

        # advance generation
        random.shuffle(population)
        for target in range(5):
            population[target] = population[-1].CloneDescendant()
        for individual in population:
            # asynchronous generations
            if random.choice([True, False]):
                individual.DepositStratum()


@pytest.mark.filterwarnings(
    "ignore:Insufficient common ranks between columns to detect common ancestry at given confidence level."
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        pytest.param(
            hstrat.HereditaryStratumOrderedStoreDict,
            marks=pytest.mark.heavy,
        ),
        hstrat.HereditaryStratumOrderedStoreList,
        pytest.param(
            hstrat.HereditaryStratumOrderedStoreTree,
            marks=pytest.mark.heavy,
        ),
    ],
)
@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_CalcRankOfMrcaBoundsWith(retention_policy, ordered_store, wrap):
    def make_bundle():
        return hstrat.HereditaryStratigraphicColumnBundle(
            {
                "test": hstrat.HereditaryStratigraphicColumn(
                    initial_stratum_annotation=0,
                    stratum_ordered_store=ordered_store,
                    stratum_retention_policy=retention_policy,
                ),
                "control": hstrat.HereditaryStratigraphicColumn(
                    initial_stratum_annotation=0,
                    stratum_ordered_store=ordered_store,
                    stratum_retention_policy=hstrat.perfect_resolution_algo.Policy(),
                ),
            }
        )

    column = make_bundle()
    frozen_copy = column.Clone()
    frozen_unrelated = make_bundle()
    population = [column.Clone() for __ in range(10)]
    forked_isolated = column.Clone()
    unrelated_isolated = make_bundle()

    for _generation in range(100):
        _ = _generation
        for f, s in it.chain(
            it.combinations(population, 2),
            zip(population, cyclify(forked_isolated)),
            zip(population, cyclify(frozen_copy)),
            zip(cyclify(forked_isolated), population),
            zip(cyclify(frozen_copy), population),
        ):
            lb, ub = hstrat.calc_rank_of_mrca_bounds_among(
                [wrap(f["test"]), wrap(s["test"])],
                prior="arbitrary",
            )
            actual_rank_of_mrca = hstrat.get_last_common_stratum_between(
                f["control"], s["control"]
            ).GetAnnotation()
            assert lb <= actual_rank_of_mrca < ub

            lb, ub = hstrat.calc_rank_of_mrca_bounds_among(
                [f["test"], s["test"], f["test"]],
                prior="arbitrary",
            )
            assert lb <= actual_rank_of_mrca < ub

        for f, s in it.chain(
            zip(population, cyclify(frozen_unrelated)),
            zip(population, cyclify(unrelated_isolated)),
            zip(cyclify(frozen_unrelated), population),
            zip(cyclify(unrelated_isolated), population),
        ):
            assert (
                hstrat.calc_rank_of_mrca_bounds_among(
                    [wrap(f["test"]), wrap(s["test"])], prior="arbitrary"
                )
                is None
            )
            assert (
                hstrat.calc_rank_of_mrca_bounds_among(
                    [wrap(f["test"]), wrap(s["test"]), wrap(f["test"])],
                    prior="arbitrary",
                )
                is None
            )
            assert (
                hstrat.calc_rank_of_mrca_bounds_among(
                    [wrap(f["test"]), wrap(s["test"])] * 3, prior="arbitrary"
                )
                is None
            )

        # advance generation
        random.shuffle(population)
        for target in range(3):
            population[target] = population[-1].CloneDescendant(
                stratum_annotation=population[-1].GetNumStrataDeposited(),
            )
        for individual in it.chain(
            iter(population),
            iterify(forked_isolated),
            iterify(unrelated_isolated),
        ):
            if random.choice([True, False]):
                individual.DepositStratum(
                    annotation=individual.GetNumStrataDeposited(),
                )


@pytest.mark.filterwarnings(
    "ignore:Insufficient common ranks between columns to detect common ancestry at given confidence level."
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        pytest.param(
            hstrat.perfect_resolution_algo.Policy(),
            marks=pytest.mark.heavy_2a,
        ),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        pytest.param(
            hstrat.HereditaryStratumOrderedStoreDict,
            marks=pytest.mark.heavy_2b,
        ),
        hstrat.HereditaryStratumOrderedStoreList,
    ],
)
@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_comparison_commutativity_synchronous(
    retention_policy,
    ordered_store,
    wrap,
):

    population = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_ordered_store=ordered_store,
            stratum_retention_policy=retention_policy,
        )
        for __ in range(10)
    ]

    for _generation in range(100):
        _ = _generation
        for first, second in it.combinations(population, 2):
            # assert commutativity
            assert (
                hstrat.calc_rank_of_mrca_bounds_among(
                    [wrap(first), wrap(second)], prior="arbitrary"
                )
                == hstrat.calc_rank_of_mrca_bounds_among(
                    [wrap(second), wrap(first)], prior="arbitrary"
                )
                == hstrat.calc_rank_of_mrca_bounds_among(
                    [wrap(second), wrap(first), wrap(first)], prior="arbitrary"
                )
                == hstrat.calc_rank_of_mrca_bounds_among(
                    [wrap(second), wrap(first), wrap(second), wrap(first)],
                    prior="arbitrary",
                )
            )

        # advance generation
        random.shuffle(population)
        for target in range(5):
            population[target] = population[-1].Clone()
        # synchronous generations
        for individual in population:
            individual.DepositStratum()


@pytest.mark.filterwarnings(
    "ignore:Insufficient common ranks between columns to detect common ancestry at given confidence level."
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        pytest.param(
            hstrat.HereditaryStratumOrderedStoreDict,
            marks=pytest.mark.heavy,
        ),
        hstrat.HereditaryStratumOrderedStoreList,
    ],
)
@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_comparison_validity(retention_policy, ordered_store, wrap):
    population = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_ordered_store=ordered_store,
            stratum_retention_policy=retention_policy,
        )
        for __ in range(10)
    ]

    for generation in range(100):
        for first, second in it.combinations(population, 2):
            lcrw = hstrat.calc_rank_of_last_retained_commonality_between(
                first, second
            )
            if lcrw is not None:
                assert 0 <= lcrw <= generation

            fdrw = hstrat.calc_rank_of_first_retained_disparity_between(
                first, second
            )
            if fdrw is not None:
                assert 0 <= fdrw <= generation

            assert hstrat.calc_rank_of_mrca_bounds_among(
                [wrap(first), wrap(second)], prior="arbitrary"
            ) in [
                (lcrw, opyt.or_value(fdrw, first.GetNumStrataDeposited())),
                None,
            ]
            assert hstrat.calc_rank_of_mrca_bounds_among(
                [wrap(first), wrap(second), wrap(first)],
                prior="arbitrary",
            ) in [
                (lcrw, opyt.or_value(fdrw, first.GetNumStrataDeposited())),
                None,
            ]
            assert hstrat.calc_rank_of_mrca_bounds_among(
                [wrap(first), wrap(second)] * 3, prior="arbitrary"
            ) in [
                (lcrw, opyt.or_value(fdrw, first.GetNumStrataDeposited())),
                None,
            ]

        # advance generations asynchronously
        random.shuffle(population)
        for target in range(5):
            population[target] = population[-1].Clone()
        for individual in population:
            if random.choice([True, False]):
                individual.DepositStratum()


@pytest.mark.filterwarnings(
    "ignore:Insufficient common ranks between columns to detect common ancestry at given confidence level."
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=10,
        ),
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 2, 8, 64],
)
@pytest.mark.parametrize(
    "confidence_level",
    [0.8, 0.95],
)
@pytest.mark.parametrize(
    "wrap",
    [
        lambda x: x,
        pytest.param(
            hstrat.col_to_specimen,
            marks=pytest.mark.heavy,
        ),
    ],
)
def test_CalcRankOfMrcaBoundsWith_narrow_shallow(
    retention_policy,
    differentia_width,
    confidence_level,
    wrap,
):

    columns = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=differentia_width,
            stratum_retention_policy=retention_policy,
            stratum_ordered_store=hstrat.HereditaryStratumOrderedStoreDict,
        )
        for __ in range(20)
    ]

    steps = list(
        range(
            columns[
                0
            ].CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
                significance_level=1 - confidence_level,
            )
            - columns[0].GetNumStrataDeposited()
        )
    )

    for step1, step2 in it.product(steps, steps):
        column1 = [col.Clone() for col in columns]
        column2 = [col.Clone() for col in columns]
        for __ in range(step1):
            for col in column1:
                col.DepositStratum()
        for i in range(step2):
            for col in column2:
                col.DepositStratum()

        for c1, c2 in zip(column1, column2):
            assert (
                hstrat.calc_rank_of_mrca_bounds_among(
                    [wrap(c1), wrap(c2)],
                    prior="arbitrary",
                    confidence_level=confidence_level,
                )
                is None
            )
            assert (
                hstrat.calc_rank_of_mrca_bounds_among(
                    [wrap(c2), wrap(c1)],
                    prior="arbitrary",
                    confidence_level=confidence_level,
                )
                is None
            )
            assert (
                hstrat.calc_rank_of_mrca_bounds_among(
                    [wrap(c2), wrap(c1), wrap(c1)],
                    prior="arbitrary",
                    confidence_level=confidence_level,
                )
                is None
            )


@pytest.mark.filterwarnings(
    "ignore:Insufficient common ranks between columns to detect common ancestry at given confidence level."
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=10,
        ),
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 2, 8, 64],
)
@pytest.mark.parametrize(
    "confidence_level",
    [0.8, 0.95],
)
@pytest.mark.parametrize(
    "mrca_rank",
    [100],
)
@pytest.mark.parametrize(
    "wrap",
    [
        lambda x: x,
        pytest.param(
            hstrat.col_to_specimen,
            marks=pytest.mark.heavy,
        ),
    ],
)
def test_CalcRankOfMrcaBoundsWith_narrow_with_mrca(
    retention_policy,
    differentia_width,
    confidence_level,
    mrca_rank,
    wrap,
):
    columns = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=differentia_width,
            stratum_retention_policy=retention_policy,
            stratum_ordered_store=hstrat.HereditaryStratumOrderedStoreDict,
        )
        for __ in range(20)
    ]

    for column in columns:
        column.DepositStrata(mrca_rank)

    steps = (0, 16, 51)

    for step1, step2 in it.product(steps, steps):
        column1 = [col.Clone() for col in columns]
        column2 = [col.Clone() for col in columns]
        for __ in range(step1):
            for col in column1:
                col.DepositStratum()
        for i in range(step2):
            for col in column2:
                col.DepositStratum()

        num_inside_bounds = 0
        num_outside_bounds = 0
        for c1, c2 in zip(column1, column2):
            assert (
                hstrat.calc_rank_of_mrca_bounds_among(
                    [wrap(c1), wrap(c2)],
                    prior="arbitrary",
                    confidence_level=confidence_level,
                )
                == hstrat.calc_rank_of_mrca_bounds_among(
                    [wrap(c2), wrap(c1)],
                    prior="arbitrary",
                    confidence_level=confidence_level,
                )
                == hstrat.calc_rank_of_mrca_bounds_among(
                    [wrap(c2), wrap(c1), wrap(c2)],
                    prior="arbitrary",
                    confidence_level=confidence_level,
                )
                == hstrat.calc_rank_of_mrca_bounds_among(
                    [wrap(c2), wrap(c1)] * 2,
                    prior="arbitrary",
                    confidence_level=confidence_level,
                )
            )
            res = hstrat.calc_rank_of_mrca_bounds_among(
                [wrap(c1), wrap(c2), wrap(c1)],
                prior="arbitrary",
                confidence_level=confidence_level,
            )

            if res is None:
                num_outside_bounds += 1
                continue

            lb, ub = res
            assert lb < ub
            assert lb >= 0
            assert ub >= 0

            num_inside_bounds += lb <= mrca_rank < ub
            num_outside_bounds += not (lb <= mrca_rank < ub)

            assert mrca_rank < ub

        num_trials = num_inside_bounds + num_outside_bounds
        assert (
            0.001
            < stats.binom.cdf(
                n=num_trials,
                p=1 - confidence_level,
                k=num_outside_bounds,
            )
            < 0.999
        )


@pytest.mark.filterwarnings(
    "ignore:Insufficient common ranks between columns to detect common ancestry at given confidence level."
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=10,
        ),
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 2, 8, 64],
)
@pytest.mark.parametrize(
    "confidence_level",
    [0.8, 0.95],
)
@pytest.mark.parametrize(
    "mrca_rank",
    [0, 100],
)
@pytest.mark.parametrize(
    "wrap",
    [
        lambda x: x,
        pytest.param(
            hstrat.col_to_specimen,
            marks=pytest.mark.heavy,
        ),
    ],
)
def test_CalcRankOfMrcaBoundsWith_narrow_no_mrca(
    retention_policy,
    differentia_width,
    confidence_level,
    mrca_rank,
    wrap,
):
    def make_column():
        return hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=differentia_width,
            stratum_retention_policy=retention_policy,
            stratum_ordered_store=hstrat.HereditaryStratumOrderedStoreDict,
        )

    columns = [make_column() for __ in range(20)]

    for column in columns:
        column.DepositStrata(mrca_rank)

    steps = (0, 16, 51)

    for step1, step2 in it.product(steps, steps):
        column1 = [make_column() for col in columns]
        column2 = [make_column() for col in columns]
        for __ in range(step1):
            for col in column1:
                col.DepositStratum()
        for i in range(step2):
            for col in column2:
                col.DepositStratum()

        num_inside_bounds = 0
        num_outside_bounds = 0
        for c1, c2 in zip(column1, column2):
            assert (
                hstrat.calc_rank_of_mrca_bounds_among(
                    [wrap(c1), wrap(c2)],
                    prior="arbitrary",
                    confidence_level=confidence_level,
                )
                == hstrat.calc_rank_of_mrca_bounds_among(
                    [wrap(c2), wrap(c1)],
                    prior="arbitrary",
                    confidence_level=confidence_level,
                )
                == hstrat.calc_rank_of_mrca_bounds_among(
                    [wrap(c2), wrap(c1), wrap(c2)],
                    prior="arbitrary",
                    confidence_level=confidence_level,
                )
                == hstrat.calc_rank_of_mrca_bounds_among(
                    [wrap(c2), wrap(c1)] * 2,
                    prior="arbitrary",
                    confidence_level=confidence_level,
                )
            )
            res = hstrat.calc_rank_of_mrca_bounds_among(
                [wrap(c1), wrap(c2), wrap(c1)],
                prior="arbitrary",
                confidence_level=confidence_level,
            )

            if res is None:
                num_inside_bounds += 1
                continue

            lb, ub = res
            assert lb < ub
            assert lb >= 0
            assert ub >= 0

            num_outside_bounds += 1

        num_trials = num_inside_bounds + num_outside_bounds
        assert (
            0.001
            < stats.binom.cdf(
                n=num_trials,
                p=1 - confidence_level,
                k=num_outside_bounds,
            )
            < 0.999
        )


@pytest.mark.filterwarnings("ignore:Empty or singleton population.")
@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_calc_rank_of_mrca_bounds_among_singletonpop(wrap):

    c1 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=1,
        stratum_retention_policy=hstrat.perfect_resolution_algo.Policy(),
    )
    for __ in range(10):
        assert (
            hstrat.calc_rank_of_earliest_detectable_mrca_among([wrap(c1)])
            is None
        )
        c1.DepositStratum()

    c1 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=64,
        stratum_retention_policy=hstrat.perfect_resolution_algo.Policy(),
    )
    for __ in range(10):
        assert (
            hstrat.calc_rank_of_earliest_detectable_mrca_among([wrap(c1)])
            is None
        )
        c1.DepositStratum()


@pytest.mark.filterwarnings("ignore:Empty or singleton population.")
def test_calc_rank_of_earliest_detectable_mrca_emptypop():

    c1 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=1,
        stratum_retention_policy=hstrat.perfect_resolution_algo.Policy(),
    )
    for __ in range(10):
        assert hstrat.calc_rank_of_earliest_detectable_mrca_among([]) is None
        c1.DepositStratum()

    c1 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=64,
        stratum_retention_policy=hstrat.perfect_resolution_algo.Policy(),
    )
    for __ in range(10):
        assert hstrat.calc_rank_of_earliest_detectable_mrca_among([]) is None
        c1.DepositStratum()


@pytest.mark.filterwarnings(
    "ignore:Insufficient common ranks between columns to detect common ancestry at given confidence level."
)
@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_calc_rank_of_mrca_bounds_among_generator(wrap):
    c1 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=1,
    )
    for __ in range(10):
        assert hstrat.calc_rank_of_mrca_bounds_among(
            [wrap(c1) for __ in range(10)],
            prior="arbitrary",
        ) == hstrat.calc_rank_of_mrca_bounds_among(
            (wrap(c1) for __ in range(10)), prior="arbitrary"
        )
        c1.DepositStratum()

    c2 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=64,
    )
    for __ in range(10):
        assert hstrat.calc_rank_of_mrca_bounds_among(
            [wrap(c2) for __ in range(10)], prior="arbitrary"
        ) == hstrat.calc_rank_of_mrca_bounds_among(
            (wrap(c2) for __ in range(10)), prior="arbitrary"
        )
        c2.DepositStratum()



import typing

from ..._detail import PolicyCouplerBase
from .._PolicySpec import PolicySpec
from .._impl import calc_provided_uncertainty


class CalcMrcaUncertaintyAbsExact:
    """Functor to provide member function implementation in Policy class."""

    def __init__(
        self: "CalcMrcaUncertaintyAbsExact",
        policy_spec: typing.Optional[PolicySpec],
    ) -> None:
        pass

    def __eq__(self: "CalcMrcaUncertaintyAbsExact", other: typing.Any) -> bool:
        return isinstance(other, self.__class__)

    def __call__(
        self: "CalcMrcaUncertaintyAbsExact",
        policy: PolicyCouplerBase,
        first_num_strata_deposited: int,
        second_num_strata_deposited: int,
        actual_rank_of_mrca: int,
    ) -> int:
        """Exactly how much uncertainty to estimate rank of MRCA?"""
        # rectify negative-indexed actual_rank_of_mrca
        if actual_rank_of_mrca is not None and actual_rank_of_mrca < 0:
            least_last_rank = min(
                first_num_strata_deposited - 1,
                second_num_strata_deposited - 1,
            )
            actual_rank_of_mrca += least_last_rank
            assert actual_rank_of_mrca >= 0

        spec = policy.GetSpec()

        least_num_strata_deposited = min(
            first_num_strata_deposited,
            second_num_strata_deposited,
        )
        least_last_rank = least_num_strata_deposited - 1

        uncertainty = calc_provided_uncertainty(
            spec.GetDepthProportionalResolution(),
            least_num_strata_deposited,
        )

        # mrca at very last rank
        if actual_rank_of_mrca == least_last_rank:
            return 0
        # haven't added enough ranks to hit resolution
        elif least_last_rank < uncertainty:
            return least_last_rank - 1
        # mrca between last regularly-spaced rank and tail rank
        elif actual_rank_of_mrca >= (
            least_last_rank - least_last_rank % uncertainty
        ):
            return (least_last_rank - 1) % uncertainty
        # mrca between two regularly-spaced ranks
        else:
            return uncertainty - 1



import itertools as it
import random

import numpy as np
import opytional as opyt
import pytest
from scipy import stats as scipy_stats

from hstrat import hstrat
from hstrat._auxiliary_lib import cmp_approx, pairwise


@pytest.mark.parametrize(
    "estimator",
    [
        "maximum_likelihood",
        "unbiased",
    ],
)
@pytest.mark.parametrize(
    "prior",
    [
        "arbitrary",
        hstrat.GeometricPrior(1.1),
        hstrat.ExponentialPrior(1.1),
        "uniform",
    ],
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=2
        ),
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 2, 8, 64],
)
def test_estimate_rank_of_mrca_between_specimen(
    estimator, prior, retention_policy, differentia_width
):
    column = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=retention_policy,
        stratum_differentia_bit_width=differentia_width,
    )
    column2 = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=retention_policy,
        stratum_differentia_bit_width=differentia_width,
    )
    column.DepositStrata(100)

    child1 = column.CloneDescendant()
    child2 = column.CloneDescendant()

    assert hstrat.estimate_rank_of_mrca_between(
        hstrat.col_to_specimen(column),
        hstrat.col_to_specimen(column2),
        estimator=estimator,
        prior=prior,
    ) == hstrat.estimate_rank_of_mrca_between(
        column,
        column2,
        estimator=estimator,
        prior=prior,
    )

    assert hstrat.estimate_rank_of_mrca_between(
        hstrat.col_to_specimen(column),
        hstrat.col_to_specimen(column),
        estimator=estimator,
        prior=prior,
    ) == hstrat.estimate_rank_of_mrca_between(
        column,
        column,
        estimator=estimator,
        prior=prior,
    )

    assert hstrat.estimate_rank_of_mrca_between(
        hstrat.col_to_specimen(column),
        hstrat.col_to_specimen(child1),
        estimator=estimator,
        prior=prior,
    ) == hstrat.estimate_rank_of_mrca_between(
        column,
        child1,
        estimator=estimator,
        prior=prior,
    )

    assert hstrat.estimate_rank_of_mrca_between(
        hstrat.col_to_specimen(child1),
        hstrat.col_to_specimen(child2),
        estimator=estimator,
        prior=prior,
    ) == hstrat.estimate_rank_of_mrca_between(
        child1,
        child2,
        estimator=estimator,
        prior=prior,
    )

    child1.DepositStrata(10)
    assert hstrat.estimate_rank_of_mrca_between(
        hstrat.col_to_specimen(child1),
        hstrat.col_to_specimen(child2),
        estimator=estimator,
        prior=prior,
    ) == hstrat.estimate_rank_of_mrca_between(
        child1,
        child2,
        estimator=estimator,
        prior=prior,
    )


@pytest.mark.filterwarnings(
    "ignore:Insufficient common ranks between columns to detect common ancestry at given confidence level."
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 8, 64],
)
@pytest.mark.parametrize(
    "estimator",
    [
        "maximum_likelihood",
        "unbiased",
    ],
)
@pytest.mark.parametrize(
    "prior",
    [
        hstrat.ArbitraryPrior(),
        hstrat.GeometricPrior(growth_factor=1.1),
        hstrat.UniformPrior(),
    ],
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
def test_comparison_commutativity_asynchronous(
    differentia_width,
    retention_policy,
    estimator,
    prior,
):
    population = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=differentia_width,
            stratum_retention_policy=retention_policy,
        )
        for __ in range(10)
    ]

    for _generation in range(30):
        _ = _generation
        for first, second in it.combinations(population, 2):
            # assert commutativity
            assert hstrat.estimate_rank_of_mrca_between(
                first,
                second,
                estimator=estimator,
                prior=prior,
            ) == hstrat.estimate_rank_of_mrca_between(
                second,
                first,
                estimator=estimator,
                prior=prior,
            )

        # advance generation
        random.shuffle(population)
        for target in range(5):
            population[target] = population[-1].CloneDescendant()
        for individual in population:
            # asynchronous generations
            if random.choice([True, False]):
                individual.DepositStratum()


@pytest.mark.filterwarnings(
    "ignore:Insufficient common ranks between columns to detect common ancestry at given confidence level."
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 8, 64],
)
@pytest.mark.parametrize(
    "estimator",
    [
        "maximum_likelihood",
        "unbiased",
    ],
)
@pytest.mark.parametrize(
    "prior",
    [
        hstrat.ArbitraryPrior(),
        hstrat.GeometricPrior(growth_factor=1.1),
        hstrat.UniformPrior(),
    ],
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=3
        ),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
def test_comparison_validity(
    differentia_width, estimator, prior, retention_policy
):
    population = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=differentia_width,
            stratum_retention_policy=retention_policy,
        )
        for __ in range(10)
    ]

    for _generation in range(50):
        _ = _generation
        for first, second in it.combinations(population, 2):

            bounds = hstrat.calc_rank_of_mrca_bounds_between(
                first, second, prior="arbitrary", confidence_level=0.49
            )
            est = hstrat.estimate_rank_of_mrca_between(
                first,
                second,
                estimator=estimator,
                prior=prior,
            )
            if bounds is not None:
                lb, ub = bounds
                assert (
                    0
                    <= est
                    < ub
                    <= min(
                        first.GetNumStrataDeposited(),
                        second.GetNumStrataDeposited(),
                    )
                )
                if estimator == "maximum_likelihood" and prior == "arbitrary":
                    assert lb <= est
            else:
                assert est is None

        # advance generations asynchronously
        random.shuffle(population)
        for target in range(5):
            population[target] = population[-1].Clone()
        for individual in population:
            if random.choice([True, False]):
                individual.DepositStratum()


@pytest.mark.filterwarnings(
    "ignore:Insufficient common ranks between columns to detect common ancestry at given confidence level."
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=3
        ),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
def test_comparison_bit_width_effects(retention_policy):
    population = [
        (
            hstrat.HereditaryStratigraphicColumn(
                stratum_differentia_bit_width=1,
                stratum_retention_policy=retention_policy,
            ),
            hstrat.HereditaryStratigraphicColumn(
                stratum_differentia_bit_width=8,
                stratum_retention_policy=retention_policy,
            ),
            hstrat.HereditaryStratigraphicColumn(
                stratum_differentia_bit_width=64,
                stratum_retention_policy=retention_policy,
            ),
        )
        for __ in range(10)
    ]

    for _generation in range(50):
        _ = _generation
        for first, second in it.combinations(population, 2):
            assert len(
                set(
                    hstrat.estimate_rank_of_mrca_between(
                        first[i],
                        second[i],
                        estimator="maximum_likelihood",
                        prior="arbitrary",
                    )
                    for i in range(3)
                )
            ) == len(
                set(
                    hstrat.calc_rank_of_first_retained_disparity_between(
                        first[i],
                        second[i],
                        confidence_level=0.49,
                    )
                    for i in range(3)
                )
            )

            for less, more in pairwise(range(3)):

                if not hstrat.calc_rank_of_last_retained_commonality_between(
                    first[less],
                    second[less],
                    confidence_level=0.49,
                ) or not hstrat.calc_rank_of_last_retained_commonality_between(
                    first[more],
                    second[more],
                    confidence_level=0.49,
                ):
                    continue

                min_len_less = min(
                    first[less].GetNumStrataDeposited(),
                    second[less].GetNumStrataDeposited(),
                )
                min_len_more = min(
                    first[more].GetNumStrataDeposited(),
                    second[more].GetNumStrataDeposited(),
                )
                assert min_len_less == min_len_more

                for estimator, prior in (
                    ("unbiased", "arbitrary"),
                    ("unbiased", "uniform"),
                    ("unbiased", hstrat.GeometricPrior(1.1)),
                ):
                    assert hstrat.estimate_rank_of_mrca_between(
                        first[less],
                        second[less],
                        estimator=estimator,
                        prior=prior,
                    ) < hstrat.estimate_rank_of_mrca_between(
                        first[more],
                        second[more],
                        estimator=estimator,
                        prior=prior,
                    ) or opyt.or_value(
                        hstrat.calc_rank_of_first_retained_disparity_between(
                            first[less],
                            second[less],
                            confidence_level=0.49,
                        ),
                        min_len_less,
                    ) > opyt.or_value(
                        hstrat.calc_rank_of_first_retained_disparity_between(
                            first[more],
                            second[more],
                            confidence_level=0.49,
                        ),
                        min_len_more,
                    )

                assert hstrat.estimate_rank_of_mrca_between(
                    first[less],
                    second[less],
                    estimator="maximum_likelihood",
                    prior="uniform",
                ) <= hstrat.estimate_rank_of_mrca_between(
                    first[more],
                    second[more],
                    estimator="maximum_likelihood",
                    prior="uniform",
                ) or opyt.or_value(
                    hstrat.calc_rank_of_first_retained_disparity_between(
                        first[less],
                        second[less],
                        confidence_level=0.49,
                    ),
                    min_len_less,
                ) > opyt.or_value(
                    hstrat.calc_rank_of_first_retained_disparity_between(
                        first[more],
                        second[more],
                        confidence_level=0.49,
                    ),
                    min_len_more,
                )

        # advance generations asynchronously
        random.shuffle(population)
        for target in range(5):
            population[target] = tuple(c.Clone() for c in population[-1])
        for individual in population:
            if random.choice([True, False]):
                for c in individual:
                    c.DepositStratum()


@pytest.mark.filterwarnings(
    "ignore:Insufficient common ranks between columns to detect common ancestry at given confidence level."
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 8, 64],
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=3
        ),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
def test_comparison_ml_vs_unbiased(differentia_width, retention_policy):
    population = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=differentia_width,
            stratum_retention_policy=retention_policy,
        )
        for __ in range(10)
    ]

    for _generation in range(50):
        _ = _generation
        for first, second in it.combinations(population, 2):

            bounds = hstrat.calc_rank_of_mrca_bounds_between(
                first, second, prior="arbitrary", confidence_level=0.49
            )
            if bounds is not None:
                assert hstrat.estimate_rank_of_mrca_between(
                    first,
                    second,
                    estimator="maximum_likelihood",
                    prior="arbitrary",
                ) >= hstrat.estimate_rank_of_mrca_between(
                    first, second, estimator="unbiased", prior="uniform"
                )

                assert (
                    cmp_approx(
                        hstrat.estimate_rank_of_mrca_between(
                            first,
                            second,
                            estimator="unbiased",
                            prior=hstrat.GeometricPrior(1.1),
                        ),
                        hstrat.estimate_rank_of_mrca_between(
                            first,
                            second,
                            estimator="unbiased",
                            prior="uniform",
                        ),
                    )
                    != -1
                )

                assert hstrat.estimate_rank_of_mrca_between(
                    first,
                    second,
                    estimator="unbiased",
                    prior="arbitrary",
                ) >= hstrat.estimate_rank_of_mrca_between(
                    first, second, estimator="unbiased", prior="uniform"
                )

                assert hstrat.estimate_rank_of_mrca_between(
                    first,
                    second,
                    estimator="maximum_likelihood",
                    prior="arbitrary",
                ) >= hstrat.estimate_rank_of_mrca_between(
                    first,
                    second,
                    estimator="maximum_likelihood",
                    prior="uniform",
                )

                assert (
                    hstrat.estimate_rank_of_mrca_between(
                        first,
                        second,
                        estimator="unbiased",
                        prior=hstrat.GeometricPrior(1.1),
                    )
                    > hstrat.estimate_rank_of_mrca_between(
                        first,
                        second,
                        estimator="unbiased",
                        prior=hstrat.GeometricPrior(1.05),
                    )
                    or not hstrat.calc_rank_of_first_retained_disparity_between(
                        first, second
                    )
                    or hstrat.calc_rank_of_last_retained_commonality_between(
                        first, second
                    )
                    == 0
                    or hstrat.calc_rank_of_first_retained_disparity_between(
                        first, second
                    )
                    == hstrat.calc_rank_of_last_retained_commonality_between(
                        first, second
                    )
                    + 1
                )

        # advance generations asynchronously
        random.shuffle(population)
        for target in range(5):
            population[target] = population[-1].Clone()
        for individual in population:
            if random.choice([True, False]):
                individual.DepositStratum()


@pytest.mark.filterwarnings(
    "ignore:Insufficient common ranks between columns to detect common ancestry at given confidence level."
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=13),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=10,
        ),
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 8, 64],
)
def test_statistical_properties_uniform_prior(
    retention_policy,
    differentia_width,
):

    err_maximum_likelihood_arbitrary = []
    err_maximum_likelihood_uniform = []
    err_unbiased = []

    common_ancestors = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=differentia_width,
            stratum_retention_policy=retention_policy,
        )
    ]
    for __ in range(113):
        common_ancestors.append(common_ancestors[-1].CloneDescendant())

    for _rep in range(8000):
        _ = _rep
        num_total = random.randrange(57, 113)
        num_together = random.randrange(num_total + 1)
        assert 0 <= num_together <= num_total
        num_alone = num_total - num_together

        left_alone = num_alone
        right_alone = num_alone

        common_ancestor = common_ancestors[num_together]
        left = common_ancestor.Clone()
        right = common_ancestor.Clone()

        left.DepositStrata(left_alone)
        right.DepositStrata(right_alone)

        err_maximum_likelihood_arbitrary.append(
            hstrat.estimate_rank_of_mrca_between(
                left, right, estimator="maximum_likelihood", prior="arbitrary"
            )
            - num_together
        )
        err_maximum_likelihood_uniform.append(
            hstrat.estimate_rank_of_mrca_between(
                left, right, estimator="maximum_likelihood", prior="uniform"
            )
            - num_together
        )
        err_unbiased.append(
            hstrat.estimate_rank_of_mrca_between(
                left, right, estimator="unbiased", prior="uniform"
            )
            - num_together
        )

    err_unbiased_ = np.array(err_unbiased)
    assert (
        np.count_nonzero(err_unbiased_) < 4
        or scipy_stats.ttest_ind(err_unbiased_, -err_unbiased_)[1] > 0.01
    )

    for err_maximum_likelihood in (
        err_maximum_likelihood_arbitrary,
        err_maximum_likelihood_uniform,
    ):
        mean_err_unbiased = np.mean(err_unbiased)
        mean_err_maximum_likelihood = np.mean(err_maximum_likelihood)

        median_abs_err_unbiased = np.percentile(np.abs(err_unbiased), 50)
        median_abs_err_maximum_likelihood = np.percentile(
            np.abs(err_maximum_likelihood), 50
        )

        if differentia_width == 1 and isinstance(
            retention_policy, hstrat.nominal_resolution_algo.Policy
        ):
            assert abs(mean_err_unbiased) < abs(mean_err_maximum_likelihood)
            assert not all(
                ml > ub
                for ml, ub in zip(
                    sorted(np.abs(err_maximum_likelihood)),
                    sorted(np.abs(err_unbiased)),
                )
            )
        elif differentia_width == 1:
            assert abs(mean_err_unbiased) < abs(mean_err_maximum_likelihood)
            assert median_abs_err_unbiased > median_abs_err_maximum_likelihood
        elif differentia_width == 8:
            _t_stat, p_value = scipy_stats.ttest_ind(
                err_unbiased, err_maximum_likelihood
            )
            _ = _t_stat
            assert (
                abs(mean_err_unbiased) <= abs(mean_err_maximum_likelihood)
                or p_value > 0.01
            )

            _w_stat, p_value = scipy_stats.wilcoxon(
                np.abs(err_unbiased), np.abs(err_maximum_likelihood)
            )
            _ = _w_stat
            assert (
                median_abs_err_unbiased >= median_abs_err_maximum_likelihood
                or p_value > 0.01
            )

        else:
            assert abs(mean_err_unbiased) <= abs(mean_err_maximum_likelihood)
            assert median_abs_err_unbiased >= median_abs_err_maximum_likelihood


@pytest.mark.filterwarnings(
    "ignore:Insufficient common ranks between columns to detect common ancestry at given confidence level."
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=7),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=10,
        ),
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 8, 64],
)
@pytest.mark.parametrize(
    "geometric_factor",
    [1.01, 1.1, 2],
)
def test_statistical_properties_geometric_prior(
    retention_policy,
    differentia_width,
    geometric_factor,
):

    err_maximum_likelihood = []
    err_unbiased = []

    common_ancestors = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=differentia_width,
            stratum_retention_policy=retention_policy,
        )
    ]
    for __ in range(113):
        common_ancestors.append(common_ancestors[-1].CloneDescendant())

    weights = np.array(
        [geometric_factor**i for i in range(113)], dtype=float
    )

    for _rep in range(8000):
        _ = _rep
        num_total = random.randrange(57, 113)

        # scrapped continuous approach to generating
        # note: e^log(b)x == b^x
        # see https://stackoverflow.com/a/40144719
        # scale = 1 / np.log(exponential_factor)
        # drawn = stats.truncexpon(b=num_total / scale, scale=scale).rvs(1)[0]
        # rounding_correction = 0.5
        # num_together = num_total - int(
        #     drawn + rounding_correction
        # )

        num_together = np.random.choice(
            np.arange(num_total + 1),
            p=weights[: num_total + 1] / weights[: num_total + 1].sum(),
        )
        assert 0 <= num_together <= num_total
        num_alone = num_total - num_together

        left_alone = num_alone
        right_alone = num_alone

        common_ancestor = common_ancestors[num_together]
        left = common_ancestor.Clone()
        right = common_ancestor.Clone()

        left.DepositStrata(left_alone)
        right.DepositStrata(right_alone)

        err_maximum_likelihood.append(
            hstrat.estimate_rank_of_mrca_between(
                left,
                right,
                estimator="maximum_likelihood",
                prior=hstrat.GeometricPrior(geometric_factor),
            )
            - num_together
        )
        err_unbiased.append(
            hstrat.estimate_rank_of_mrca_between(
                left,
                right,
                estimator="unbiased",
                prior=hstrat.GeometricPrior(geometric_factor),
            )
            - num_together
        )

    err_unbiased_ = np.array(err_unbiased)

    assert (
        np.count_nonzero(err_unbiased_) < 4
        or scipy_stats.ttest_1samp(err_unbiased_, 0)[1] > 0.01
    )

    mean_err_unbiased = np.mean(err_unbiased)
    mean_err_maximum_likelihood = np.mean(err_maximum_likelihood)

    assert not all(
        ml > ub
        for ml, ub in zip(
            sorted(np.abs(err_maximum_likelihood)),
            sorted(np.abs(err_unbiased)),
        )
    )

    if differentia_width == 1:
        assert abs(mean_err_unbiased) < abs(mean_err_maximum_likelihood)
    else:
        assert abs(mean_err_unbiased) <= abs(mean_err_maximum_likelihood)


@pytest.mark.parametrize(
    "common_ancestor",
    [
        hstrat.HereditaryStratigraphicColumn(
            stratum_retention_policy=hstrat.nominal_resolution_algo.Policy(),
        ),
        hstrat.HereditaryStratigraphicColumn(
            stratum_retention_policy=hstrat.nominal_resolution_algo.Policy(),
        ).CloneNthDescendant(5),
    ],
)
@pytest.mark.parametrize(
    "estimator",
    ["maximum_likelihood", "unbiased"],
)
@pytest.mark.parametrize(
    "prior",
    ["arbitrary", "uniform"],
)
def test_nominal_retention_policy(common_ancestor, estimator, prior):
    assert (
        hstrat.estimate_rank_of_mrca_between(
            common_ancestor, common_ancestor, estimator=estimator, prior=prior
        )
        == common_ancestor.GetNumStrataDeposited() - 1
    )

    assert (
        hstrat.calc_rank_of_first_retained_disparity_between(
            common_ancestor,
            common_ancestor.CloneNthDescendant(7),
        )
        == common_ancestor.GetNumStrataDeposited()
    )

    assert (
        hstrat.estimate_rank_of_mrca_between(
            common_ancestor,
            common_ancestor.CloneNthDescendant(7),
            estimator=estimator,
            prior=prior,
        )
        == (common_ancestor.GetNumStrataDeposited() - 1) / 2
    )

    assert (
        hstrat.estimate_rank_of_mrca_between(
            common_ancestor.CloneNthDescendant(2),
            common_ancestor.CloneNthDescendant(7),
            estimator=estimator,
            prior=prior,
        )
        == (common_ancestor.GetNumStrataDeposited() - 1 + 2) / 2
    )

    assert (
        hstrat.estimate_rank_of_mrca_between(
            common_ancestor.CloneNthDescendant(7),
            common_ancestor.CloneNthDescendant(7),
            estimator=estimator,
            prior=prior,
        )
        == (common_ancestor.GetNumStrataDeposited() + 7 - 1 - 1) / 2
    )


@pytest.mark.parametrize(
    "differentia_width",
    [1, 8, 64],
)
@pytest.mark.parametrize(
    "policy",
    [
        hstrat.fixed_resolution_algo.Policy(3),
        hstrat.recency_proportional_resolution_algo.Policy(1),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.perfect_resolution_algo.Policy(),
    ],
)
@pytest.mark.parametrize(
    "estimator",
    ["maximum_likelihood", "unbiased"],
)
@pytest.mark.parametrize(
    "prior",
    ["arbitrary", "uniform"],
)
def test_artifact_types_equiv(differentia_width, policy, estimator, prior):
    common_ancestor = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    ).CloneNthDescendant(7)
    c1 = common_ancestor.CloneNthDescendant(4)
    c2 = common_ancestor.CloneNthDescendant(9)
    c_x = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    ).CloneNthDescendant(7)
    c_y = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    )

    for a, b in it.product(
        [common_ancestor, c1, c2, c_x, c_y],
        [common_ancestor, c1, c2, c_x, c_y],
    ):
        assert hstrat.estimate_rank_of_mrca_between(
            hstrat.col_to_specimen(a),
            hstrat.col_to_specimen(b),
            estimator,
            prior,
        ) == hstrat.estimate_rank_of_mrca_between(a, b, estimator, prior)



import pandas as pd

from ._alifestd_mark_roots import alifestd_mark_roots


def alifestd_mark_oldest_root(
    phylogeny_df: pd.DataFrame, mutate: bool = False
) -> pd.DataFrame:
    """Point all other roots to oldest root, measured by lowest `origin_time`
    (if available) or otherwise lowest `id`.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """
    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df = alifestd_mark_roots(phylogeny_df, mutate=True)
    if len(phylogeny_df) <= 1:
        phylogeny_df["is_oldest_root"] = True
        return phylogeny_df

    phylogeny_df.reset_index(drop=True, inplace=True)
    idxmin: int
    if "origin_time" in phylogeny_df.columns:
        idxmin = (
            phylogeny_df.loc[phylogeny_df["is_root"]]
            .sort_values(by=["origin_time", "id"])
            .index[0]
        )
    else:
        idxmin = phylogeny_df.loc[phylogeny_df["is_root"], "id"].idxmin()

    phylogeny_df["is_oldest_root"] = False
    phylogeny_df.loc[idxmin, "is_oldest_root"] = True

    return phylogeny_df



import numpy as np
import pytest

from hstrat.hstrat import fixed_resolution_algo


@pytest.mark.parametrize(
    "fixed_resolution",
    [
        1,
        2,
        3,
        7,
        42,
        100,
    ],
)
@pytest.mark.parametrize(
    "time_sequence",
    [
        range(10**2),
        (i for i in range(10**2) for __ in range(2)),
        np.random.default_rng(1).integers(
            low=0,
            high=10**3,
            size=10**2,
        ),
    ],
)
def test_policy_consistency(fixed_resolution, time_sequence):
    policy = fixed_resolution_algo.Policy(fixed_resolution)
    spec = policy.GetSpec()
    instance = fixed_resolution_algo.CalcNumStrataRetainedExact(spec)
    for num_strata_deposited in time_sequence:
        policy_requirement = len(
            [
                *policy.IterRetainedRanks(
                    num_strata_deposited,
                )
            ]
        )
        for which in (
            instance,
            fixed_resolution_algo.CalcNumStrataRetainedExact(spec),
        ):
            assert (
                which(
                    policy,
                    num_strata_deposited,
                )
                == policy_requirement
            )


@pytest.mark.parametrize(
    "fixed_resolution",
    [
        1,
        2,
        3,
        7,
        42,
        100,
    ],
)
def test_eq(fixed_resolution):
    policy = fixed_resolution_algo.Policy(fixed_resolution)
    spec = policy.GetSpec()
    instance = fixed_resolution_algo.CalcNumStrataRetainedExact(spec)

    assert instance == instance
    assert instance == fixed_resolution_algo.CalcNumStrataRetainedExact(
        spec,
    )
    assert instance is not None



import networkx as nx

import hstrat.phylogenetic_inference.tree._impl as impl


def test_find_chronological_root_with_simple_tree():
    graph = nx.Graph()
    graph.add_edge("A", "B", length=8)
    graph.add_edge("B", "C", length=2)
    graph.add_edge("C", "D", length=1)
    graph.add_edge("C", "E", length=11)
    leaf_node_origin_times = {"D": 52, "E": 62}

    for node in graph.nodes:
        graph.nodes[node]["taxon_label"] = node

    root = impl.find_chronological_roots(graph, leaf_node_origin_times)
    assert root == ["A"]


def test_find_chronological_root_with_empty_tree():
    graph = nx.Graph()
    leaf_node_origin_times = {}
    roots = impl.find_chronological_roots(graph, leaf_node_origin_times)
    assert roots == []


def test_find_chronological_root_with_singleton():
    graph = nx.Graph()
    graph.add_node("Q")
    leaf_node_origin_times = {"Q": 0}
    for node in graph.nodes:
        graph.nodes[node]["taxon_label"] = node

    roots = impl.find_chronological_roots(graph, leaf_node_origin_times)
    assert roots == ["Q"]


def test_find_chronological_root_with_disjoint_tree1():
    graph = nx.Graph()
    graph.add_edge("A", "B", length=8)
    graph.add_edge("B", "C", length=2)
    graph.add_edge("C", "D", length=1)
    graph.add_edge("C", "E", length=11)
    graph.add_node("Q")
    leaf_node_origin_times = {"D": 52, "E": 62, "Q": 17}

    for node in graph.nodes:
        graph.nodes[node]["taxon_label"] = node

    root = impl.find_chronological_roots(graph, leaf_node_origin_times)
    assert root == ["A", "Q"]


def test_find_chronological_root_with_disjoint_tree2():
    graph = nx.Graph()
    graph.add_edge("A", "B", length=8)
    graph.add_edge("B", "C", length=2)
    graph.add_edge("C", "D", length=1)
    graph.add_edge("C", "E", length=11)
    graph.add_edge("Q", "R", length=12)
    leaf_node_origin_times = {"D": 52, "E": 62, "Q": 17}

    for node in graph.nodes:
        graph.nodes[node]["taxon_label"] = node

    root = impl.find_chronological_roots(graph, leaf_node_origin_times)
    assert root == ["A", "R"]



import pytest

from hstrat import hstrat


@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_calc_probability_differentia_collision_between(wrap):
    assert (
        hstrat.calc_probability_differentia_collision_between(
            wrap(
                hstrat.HereditaryStratigraphicColumn(
                    stratum_differentia_bit_width=1
                )
            ),
            wrap(
                hstrat.HereditaryStratigraphicColumn(
                    stratum_differentia_bit_width=1
                )
            ),
        )
        == 0.5
    )
    assert (
        hstrat.calc_probability_differentia_collision_between(
            wrap(
                hstrat.HereditaryStratigraphicColumn(
                    stratum_differentia_bit_width=2
                )
            ),
            wrap(
                hstrat.HereditaryStratigraphicColumn(
                    stratum_differentia_bit_width=2
                )
            ),
        )
        == 0.25
    )
    assert (
        hstrat.calc_probability_differentia_collision_between(
            wrap(
                hstrat.HereditaryStratigraphicColumn(
                    stratum_differentia_bit_width=3
                )
            ),
            wrap(
                hstrat.HereditaryStratigraphicColumn(
                    stratum_differentia_bit_width=3
                )
            ),
        )
        == 0.125
    )
    assert (
        hstrat.calc_probability_differentia_collision_between(
            wrap(
                hstrat.HereditaryStratigraphicColumn(
                    stratum_differentia_bit_width=8
                )
            ),
            wrap(
                hstrat.HereditaryStratigraphicColumn(
                    stratum_differentia_bit_width=8
                )
            ),
        )
        == 1.0 / 256.0
    )



import typing
import warnings

import opytional as opyt

from ..._auxiliary_lib import HereditaryStratigraphicArtifact
from ...juxtaposition import (
    calc_definitive_min_ranks_since_first_retained_disparity_with,
    calc_ranks_since_last_retained_commonality_with,
)
from ._calc_rank_of_earliest_detectable_mrca_between import (
    calc_rank_of_earliest_detectable_mrca_between,
)
from ._does_have_any_common_ancestor import does_have_any_common_ancestor


def calc_ranks_since_mrca_bounds_with(
    focal: HereditaryStratigraphicArtifact,
    other: HereditaryStratigraphicArtifact,
    prior: typing.Literal["arbitrary"],
    confidence_level: float = 0.95,
) -> typing.Optional[typing.Tuple[int, int]]:
    """How many generations have elapsed since MRCA?

    Calculate bounds on estimate for the number of depositions elapsed
    along focal column's line of descent since the most recent common
    ancestor with other.

    Parameters
    ----------
    prior : {"arbitrary"}
        Prior probability density distribution over possible generations of the
        MRCA.

        Currently only "arbitrary" supported.
    confidence_level : float, optional
        With what probability should the true rank of the MRCA fall
        within the calculated bounds? Default 0.95.

    Returns
    -------
    (int, int), optional
        Inclusive lower bound and then exclusive upper bound on estimate or
        None if no common ancestor between focal and other can be resolved
        with sufficient confidence. (Sufficient confidence depends on
        bound_type.)

    See Also
    --------
    calc_ranks_since_mrca_uncertainty_between :
        Wrapper to report uncertainty of calculated bounds.
    calc_ranks_since_earliest_detectable_mrca_between :
        Could any MRCA be detected between focal and other? How many ranks
        have elapsed since the earliest MRCA that could be reliably
        detected?
    calc_ranks_since_mrca_bounds_provided_confidence_level :
        With what actual confidence (i.e., more than requested) is the true
        rank of the MRCA captured within the calculated bounds?
    does_definitively_have_no_common_anestor :
        Does the hereditary stratigraphic record definitively prove that first
        and second could not possibly share a common ancestor?

    Notes
    -----
    The true number of ranks since the MRCA is guaranteed to never fall
    below the bounds but may fall above.

    An alternate approach could be to construct the bounds such that the
    true number of ranks since the MRCA will fall above or below the bounds
    with equal probability. This would involve setting the confidence level
    for calculating the first disparity with other to significance_level/2
    and the confidence level for calculaing the last comonality with other
    to 1 - significance_level/2. This means the confidence level applied to
    calculating the first disparity with other would always be <= 0.5.
    However, shifting the calculated first disparity with other above
    the definitive min first retained disparity requires confidence level
    >= 0.5. So, in practice such a symmetric approach would only result in
    the upper bound being shifted upward. For this reason, it is no longer
    provided as an option.

    In the absence of evidence to the contrary (i.e., more common
    strata than spurious differentia collisions alone could plausibly
    cause), this method assumes no common ancestry between focal and other,
    returning None. This means that if few enough common ranks are shared
    between focal and other (and the differentia bit with is small enough),
    it may not be possible to detect any common ancestry after accounting
    for the possibility of spurious differentia collisions (even if common
    ancestry did exist). So, calls to this method would always return None.
    Likewise, MRCAs at very early ranks may not be able to be reliably
    detected due to insufficient evidence. This can lead to cases where
    columns with true common ancestry have MRCA bounds estimated as None at
    much higher than the expected failure rate at the given confidence
    level. Note that with sufficient differentia bit width (i.e., so that
    even one collision is implausible at the given confidence level) this
    issue does not occur. Use CalcRanksSinceEarliestDetectableMrcaWith to
    determine the earliest rank at which an MRCA could be reliably detected
    between focal and other.
    """
    if prior != "arbitrary":
        raise NotImplementedError
    assert 0.0 <= confidence_level <= 1.0

    if (
        calc_rank_of_earliest_detectable_mrca_between(
            focal,
            other,
            confidence_level=confidence_level,
        )
        is None
    ):
        warnings.warn(
            "Insufficient common ranks between columns to detect common "
            "ancestry at given confidence level."
        )

    if does_have_any_common_ancestor(
        focal,
        other,
        confidence_level=confidence_level,
    ):
        since_first_disparity = (
            calc_definitive_min_ranks_since_first_retained_disparity_with(
                focal,
                other,
            )
        )

        lb_exclusive = opyt.or_value(since_first_disparity, -1)
        lb_inclusive = lb_exclusive + 1

        since_last_commonality = (
            calc_ranks_since_last_retained_commonality_with(
                focal,
                other,
                confidence_level=confidence_level,
            )
        )
        assert since_last_commonality is not None
        ub_inclusive = since_last_commonality
        ub_exclusive = ub_inclusive + 1

        return (lb_inclusive, ub_exclusive)
    else:
        return None



import unittest

from hstrat import hstrat
from hstrat._auxiliary_lib import release_cur_mpl_fig


class TestPolicyPanelPlot(unittest.TestCase):

    # tests can run independently
    _multiprocess_can_split_ = True

    def test(self):
        for policy in [
            hstrat.fixed_resolution_algo.Policy(10),
            hstrat.nominal_resolution_algo.Policy(),
            hstrat.perfect_resolution_algo.Policy(),
        ]:
            hstrat.policy_panel_plot(policy, 100, do_show=False)
            release_cur_mpl_fig()
            hstrat.policy_panel_plot(policy, 10, do_show=False)
            release_cur_mpl_fig()


if __name__ == "__main__":
    unittest.main()



import anytree
import networkx as nx
import pytest

from hstrat import hstrat
from hstrat._auxiliary_lib import random_tree, seed_random
import hstrat.phylogenetic_inference.tree._impl as impl
import hstrat.phylogenetic_inference.tree.trie_postprocess._detail as detail


def test_base_class():
    assert issubclass(
        hstrat.AssignOriginTimeSampleNaiveTriePostprocessor,
        detail.TriePostprocessorBase,
    )


def test_assign_trie_origin_times_naive_single_leaf():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    leaf = impl.TrieLeafNode(parent=inner, taxon_label="A")
    root = hstrat.AssignOriginTimeSampleNaiveTriePostprocessor()(
        root, p_differentia_collision=0.5, mutate=True
    )
    assert leaf.origin_time == 0
    assert inner.origin_time == 0
    assert root.origin_time == 0


def test_assign_trie_origin_times_naive_two_leaves():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    leaf1 = impl.TrieLeafNode(parent=inner, taxon_label="A")
    leaf2 = impl.TrieLeafNode(parent=inner, taxon_label="B")
    root = hstrat.AssignOriginTimeSampleNaiveTriePostprocessor()(
        root, p_differentia_collision=0.5, mutate=True
    )
    assert root.origin_time == 0
    assert inner.origin_time == 0
    assert leaf1.origin_time == 0
    assert leaf2.origin_time == 0


def test_assign_trie_origin_times_naive1():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner1 = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    inner2 = impl.TrieInnerNode(rank=1, differentia=10, parent=inner1)
    leaf = impl.TrieLeafNode(parent=inner2, taxon_label="A")
    root = hstrat.AssignOriginTimeSampleNaiveTriePostprocessor()(
        root, p_differentia_collision=0.5, mutate=True
    )
    assert root.origin_time == 0
    assert inner1.origin_time == 0
    assert 0 <= inner2.origin_time <= 1
    assert leaf.origin_time == 1


def test_assign_trie_origin_times_naive2():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner1 = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    inner2 = impl.TrieInnerNode(rank=2, differentia=10, parent=inner1)
    leaf = impl.TrieLeafNode(parent=inner2, taxon_label="A")
    root = hstrat.AssignOriginTimeSampleNaiveTriePostprocessor()(
        root, p_differentia_collision=0.5, mutate=True
    )
    assert root.origin_time == 0
    assert 0 <= inner1.origin_time <= 0.5
    assert 0.5 <= inner2.origin_time <= 2
    assert leaf.origin_time == 2


def test_assign_trie_origin_times_naive3():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner1 = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    inner2 = impl.TrieInnerNode(rank=3, differentia=10, parent=inner1)
    leaf = impl.TrieLeafNode(parent=inner2, taxon_label="A")
    root = hstrat.AssignOriginTimeSampleNaiveTriePostprocessor()(
        root, p_differentia_collision=0.5, mutate=True
    )
    assert root.origin_time == 0
    assert 0 <= inner1.origin_time <= 2
    assert inner2.origin_time == 3
    assert leaf.origin_time == 3


def test_assign_trie_origin_times_naive_complex():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner1 = impl.TrieInnerNode(rank=0, differentia=0, parent=root)

    inner2 = impl.TrieInnerNode(rank=4, differentia=10, parent=inner1)
    inner3 = impl.TrieInnerNode(rank=7, differentia=7, parent=inner2)

    inner2a = impl.TrieInnerNode(rank=1, differentia=1, parent=inner1)

    leaf3_a = impl.TrieLeafNode(parent=inner3, taxon_label="leaf3a_a")

    leaf2a_a = impl.TrieLeafNode(parent=inner2a, taxon_label="leaf2a_a")
    leaf2a_b = impl.TrieLeafNode(parent=inner2a, taxon_label="leaf2a_b")

    root = hstrat.AssignOriginTimeNaiveTriePostprocessor()(
        root, p_differentia_collision=0.5, mutate=True
    )

    assert root.origin_time == 0
    assert inner1.origin_time == 0
    assert 4 <= inner2.origin_time <= 6
    assert inner3.origin_time == 7
    assert inner2a.origin_time == 1
    assert leaf2a_a.origin_time == 1
    assert leaf2a_b.origin_time == 1
    assert leaf3_a.origin_time == 7


def test_assign_trie_origin_times_naive_assigned_property():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    leaf = impl.TrieLeafNode(parent=inner, taxon_label="A")
    root = hstrat.AssignOriginTimeSampleNaiveTriePostprocessor(
        assigned_property="blueberry"
    )(root, p_differentia_collision=0.5, mutate=True)
    assert leaf.blueberry == 0
    assert inner.blueberry == 0
    assert root.blueberry == 0


def test_assign_trie_origin_times_naive_mutate():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    leaf = impl.TrieLeafNode(parent=inner, taxon_label="A")
    root_ = hstrat.AssignOriginTimeSampleNaiveTriePostprocessor(
        assigned_property="blueberry"
    )(root, p_differentia_collision=0.5, mutate=False)
    assert not hasattr(leaf, "blueberry")
    assert not hasattr(inner, "blueberry")
    assert not hasattr(root, "blueberry")

    assert hasattr(root_, "blueberry")

    root = hstrat.AssignOriginTimeSampleNaiveTriePostprocessor(
        prior=hstrat.ArbitraryPrior(),
        assigned_property="blueberry",
    )(root, p_differentia_collision=0.5, mutate=True)
    assert str(anytree.RenderTree(root).by_attr("blueberry")) == str(
        anytree.RenderTree(root_).by_attr("blueberry")
    )


@pytest.mark.parametrize("tree_size", [300])
@pytest.mark.parametrize("differentia_width", [64])
@pytest.mark.parametrize("tree_seed", range(10))
@pytest.mark.parametrize(
    "retention_policy", [hstrat.perfect_resolution_algo.Policy()]
)
def test_perfect_rank_equivalence(
    tree_seed, tree_size, differentia_width, retention_policy
):
    seed_random(tree_seed)
    nx_tree = random_tree(n=tree_size, seed=tree_seed, create_using=nx.DiGraph)

    extant_population = hstrat.descend_template_phylogeny_networkx(
        nx_tree,
        seed_column=hstrat.HereditaryStratigraphicColumn(
            stratum_retention_policy=retention_policy,
            stratum_differentia_bit_width=differentia_width,
        ),
    )

    rank_df, naive_df = hstrat.build_tree_trie_ensemble(
        extant_population,
        trie_postprocessors=[
            hstrat.AssignOriginTimeNodeRankTriePostprocessor(),
            hstrat.AssignOriginTimeSampleNaiveTriePostprocessor(),
        ],
    )

    assert (rank_df["origin_time"] == naive_df["origin_time"]).all()



import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_chronological_sort,
    alifestd_is_chronologically_sorted,
    alifestd_make_empty,
)


def test_chronological_sort_empty():
    df = alifestd_make_empty()
    df["origin_time"] = []
    df = alifestd_chronological_sort(df)
    assert alifestd_is_chronologically_sorted(df)


@pytest.mark.parametrize("mutate", [True, False])
def test_chronological_sort(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            "origin_time": [0, 1, 2, 2, 6, 5],
        }
    )
    original_df = phylogeny_df.copy()
    expected_result_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 5, 44],
            "origin_time": [0, 1, 2, 2, 5, 6],
        }
    )
    result_df = alifestd_chronological_sort(phylogeny_df, mutate=mutate)

    assert alifestd_is_chronologically_sorted(result_df)
    assert expected_result_df.equals(result_df)

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_chronological_sort_nop(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            "origin_time": [0, 1, 2, 2, 5, 6],
        }
    )
    original_df = phylogeny_df.copy()
    expected_result_df = original_df
    result_df = alifestd_chronological_sort(phylogeny_df, mutate=mutate)

    assert alifestd_is_chronologically_sorted(result_df)
    assert expected_result_df.equals(result_df)

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
@pytest.mark.parametrize("how", ["origin_time", "destruction_time"])
def test_chronological_sort_how(mutate: bool, how: str):
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            how: [0, 1, 2, 2, 6, 5],
        }
    )
    original_df = phylogeny_df.copy()
    expected_result_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 5, 44],
            how: [0, 1, 2, 2, 5, 6],
        }
    )
    result_df = alifestd_chronological_sort(
        phylogeny_df, how=how, mutate=mutate
    )

    assert alifestd_is_chronologically_sorted(result_df, how=how)
    assert expected_result_df.equals(result_df)

    if not mutate:
        assert original_df.equals(phylogeny_df)



import numpy as np

from hstrat._auxiliary_lib import iter_monotonic_equivalencies_reverse


def test_empty_arrays():
    first = np.array([])
    second = np.array([])
    res = [*iter_monotonic_equivalencies_reverse(first, second)]
    assert res == [*reversed([])]
    for i, j in res:
        assert first[i] == second[j]


def test_one_element_arrays():
    first = np.array([1])
    second = np.array([1])
    res = [*iter_monotonic_equivalencies_reverse(first, second)]
    assert res == [*reversed([(0, 0)])]
    for i, j in res:
        assert first[i] == second[j]


def test_identical_arrays():
    first = np.array([1, 2, 3, 4, 5])
    second = np.array([1, 2, 3, 4, 5])
    res = [*iter_monotonic_equivalencies_reverse(first, second)]
    assert res == [*reversed([(0, 0), (1, 1), (2, 2), (3, 3), (4, 4)])]
    for i, j in res:
        assert first[i] == second[j]


def test_non_overlapping_arrays():
    first = np.array([1, 3, 5, 7, 9])
    second = np.array([2, 4, 6, 8, 10])
    res = [*iter_monotonic_equivalencies_reverse(first, second)]
    assert res == [*reversed([])]
    for i, j in res:
        assert first[i] == second[j]


def test_some_overlapping_arrays():
    first = np.array([1, 2, 3, 4, 5])
    second = np.array([2, 4, 6, 8, 10])
    res = [*iter_monotonic_equivalencies_reverse(first, second)]
    assert res == [*reversed([(1, 0), (3, 1)])]
    for i, j in res:
        assert first[i] == second[j]


def test_non_decreasing_arrays():
    first = np.array([1, 2, 3, 3, 3, 4, 5, 7])
    second = np.array([2, 3, 3, 4, 4, 4, 6, 7])
    res = [*iter_monotonic_equivalencies_reverse(first, second)]
    assert res == [*reversed([(1, 0), (3, 1), (4, 2), (5, 5), (7, 7)])]
    for i, j in res:
        assert first[i] == second[j]


def test_start_parameter1():
    first = np.array([1, 2, 3, 4, 5])
    second = np.array([2, 4, 6, 8, 10])
    res = [*iter_monotonic_equivalencies_reverse(first, second, start=(1, 0))]
    assert res == [*reversed([(1, 0)])]
    for i, j in res:
        assert first[i] == second[j]


def test_start_parameter2():
    first = np.array([1, 2, 3, 4, 5])
    second = np.array([2, 4, 6, 8, 10])
    res = [*iter_monotonic_equivalencies_reverse(first, second, start=(0, 0))]
    assert res == [*reversed([])]
    for i, j in res:
        assert first[i] == second[j]


def test_start_parameter3():
    first = np.array([1, 2, 3, 4, 5])
    second = np.array([2, 4, 6, 8, 10])
    res = [
        *iter_monotonic_equivalencies_reverse(first, second, start=(-3, -1))
    ]
    assert res == [*reversed([(1, 0)])]
    for i, j in res:
        assert first[i] == second[j]



import typing

from ...frozen_instrumentation import HereditaryStratigraphicSpecimen
from ._iter_mutual_ranks import iter_mutual_ranks


def iter_ranks_of_retained_commonality_between(
    first: HereditaryStratigraphicSpecimen,
    second: HereditaryStratigraphicSpecimen,
) -> typing.Iterator[int]:
    """Iterate over ranks with matching strata between columns in ascending
    order."""
    for rank, comp in iter_mutual_ranks(first, second, compare=True):
        if comp:
            yield rank
        else:
            return



import typing

import numpy as np
import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_find_root_ids,
    alifestd_make_empty,
    alifestd_try_add_ancestor_id_col,
)


@pytest.mark.parametrize(
    "apply", [alifestd_try_add_ancestor_id_col, lambda x: x]
)
def test_no_root(apply: typing.Callable):
    assert np.array_equal(
        alifestd_find_root_ids(apply(alifestd_make_empty())),
        np.array([]),
    )


@pytest.mark.parametrize(
    "apply", [alifestd_try_add_ancestor_id_col, lambda x: x]
)
def test_singleton(apply: typing.Callable):
    phylogeny_df = pd.DataFrame(
        {
            "id": [0],
            "ancestor_list": ["[none]"],
        }
    )
    phylogeny_df = apply(phylogeny_df)
    assert np.array_equal(alifestd_find_root_ids(phylogeny_df), np.array([0]))


@pytest.mark.parametrize(
    "apply", [alifestd_try_add_ancestor_id_col, lambda x: x]
)
def test_alifestd_find_root_ids_returns_expected_output_with_single_root1(
    apply: typing.Callable,
):
    phylogeny_df = pd.DataFrame(
        {
            "id": [1, 2, 3],
            "ancestor_list": ["[none]", "[1]", "[2]"],
        }
    )
    phylogeny_df = apply(phylogeny_df)

    expected_output = np.array([1])
    output = alifestd_find_root_ids(phylogeny_df)
    assert np.array_equal(output, expected_output)


@pytest.mark.parametrize(
    "apply", [alifestd_try_add_ancestor_id_col, lambda x: x]
)
def test_alifestd_find_root_ids_returns_expected_output_with_multiple_roots1(
    apply: typing.Callable,
):
    phylogeny_df = pd.DataFrame(
        {
            "id": [1, 2, 3, 4],
            "ancestor_list": ["[none]", "[none]", "[1]", "[2]"],
        }
    )
    phylogeny_df = apply(phylogeny_df)
    expected_output = np.array([1, 2])
    output = alifestd_find_root_ids(phylogeny_df)
    assert np.array_equal(output, expected_output)


def test_alifestd_find_root_ids_returns_expected_output_with_single_root2():
    phylogeny_df = pd.DataFrame(
        {
            "id": [1, 2, 3],
            "ancestor_list": ["[none]", "[1]", "[2, 1]"],
        }
    )
    expected_output = np.array([1])
    output = alifestd_find_root_ids(phylogeny_df)
    assert np.array_equal(output, expected_output)


def test_alifestd_find_root_ids_returns_expected_output_with_multiple_roots2():
    phylogeny_df = pd.DataFrame(
        {
            "id": [1, 2, 3, 4],
            "ancestor_list": ["[none]", "[none]", "[1]", "[2, 1]"],
        }
    )
    expected_output = np.array([1, 2])
    output = alifestd_find_root_ids(phylogeny_df)
    assert np.array_equal(output, expected_output)


@pytest.mark.parametrize(
    "apply", [alifestd_try_add_ancestor_id_col, lambda x: x]
)
def test_alifestd_find_root_ids_does_not_mutate_input_dataframe(
    apply: typing.Callable,
):
    phylogeny_df = pd.DataFrame(
        {
            "id": [1, 2, 3],
            "ancestor_list": ["[none]", "[1]", "[2]"],
        }
    )
    phylogeny_df = apply(phylogeny_df)
    input_df = phylogeny_df.copy()
    alifestd_find_root_ids(input_df)
    assert phylogeny_df.equals(input_df)


@pytest.mark.parametrize(
    "apply", [alifestd_try_add_ancestor_id_col, lambda x: x]
)
def test_alifestd_find_root_ids_returns_expected_output_with_single_root3(
    apply: typing.Callable,
):
    phylogeny_df = pd.DataFrame(
        {
            "id": [1, 2, 3],
            "ancestor_list": [[None], [1], [2]],
        }
    )
    phylogeny_df = apply(phylogeny_df)
    expected_output = np.array([1])
    output = alifestd_find_root_ids(phylogeny_df)
    assert np.array_equal(output, expected_output)


@pytest.mark.parametrize(
    "apply", [alifestd_try_add_ancestor_id_col, lambda x: x]
)
def test_alifestd_find_root_ids_returns_expected_output_with_multiple_roots3(
    apply: typing.Callable,
):
    phylogeny_df = pd.DataFrame(
        {
            "id": [1, 2, 3, 4],
            "ancestor_list": [[None], [None], [1], [2]],
        }
    )
    phylogeny_df = apply(phylogeny_df)
    expected_output = np.array([1, 2])
    output = alifestd_find_root_ids(phylogeny_df)
    assert np.array_equal(output, expected_output)


@pytest.mark.parametrize(
    "apply", [alifestd_try_add_ancestor_id_col, lambda x: x]
)
def test_alifestd_find_root_ids_does_not_mutate_input_dataframe3(
    apply: typing.Callable,
):
    phylogeny_df = pd.DataFrame(
        {
            "id": [1, 2, 3],
            "ancestor_list": [[None], [1], [2]],
        }
    )
    phylogeny_df = apply(phylogeny_df)
    input_df = phylogeny_df.copy()
    alifestd_find_root_ids(input_df)
    assert phylogeny_df.equals(input_df)



import pandas as pd

from ._alifestd_make_ancestor_list_col_polars import (
    alifestd_make_ancestor_list_col_polars,
)
from ._delegate_polars_implementation import (
    Series_T,
    delegate_polars_implementation,
)


@delegate_polars_implementation(alifestd_make_ancestor_list_col_polars)
def alifestd_make_ancestor_list_col(
    ids: Series_T,
    ancestor_ids: Series_T,
    root_ancestor_token: str = "none",
) -> Series_T:
    """Translate a column of integer ancestor id values into alife standard
    `ancestor_list` representation.

    The option `root_ancestor_token` will be sandwiched in brackets to create
    the ancestor list entry for genesis organisms. For example, the token
    "None" will yield the entry "[None]" and the token "" will yield the entry
    "[]". Default "none".
    """

    assert isinstance(ids, pd.Series) and isinstance(ancestor_ids, pd.Series)
    res = ancestor_ids.map("[{!s}]".format).astype(str)  # specify for empty
    res[ids == ancestor_ids] = f"[{root_ancestor_token}]"

    return res



import typing

from deprecated.sphinx import deprecated

from ._unpack_differentiae_str import unpack_differentiae_str


@deprecated(
    version="1.8.0",
    reason="Use unpack_differentiae_str instead.",
)
def unpack_differentiae(
    packed_differentiae: str,
    differentia_bit_width: int,
) -> typing.Iterator[int]:
    """Unpack a compact, concatenated base 64 representation into
    a sequence with each element represented as a distinct integer."""
    yield from unpack_differentiae_str(
        packed_differentiae, differentia_bit_width
    )



from ._auxiliary_lib import get_hstrat_version

if __name__ == "__main__":
    print(f"hstrat v{get_hstrat_version()}")
    print()
    print("Available commands (stabilized API):")
    print("$ python3 -m hstrat.dataframe.surface_unpack_reconstruct")
    print("$ python3 -m hstrat.dataframe.surface_postprocess_trie")
    print()
    print("Available commands (experimental API):")
    print("$ python3 -m hstrat._auxiliary_lib._alifestd_as_newick_asexual")
    print(
        "$ python3 -m hstrat._auxiliary_lib._alifestd_try_add_ancestor_list_col"
    )
    print()
    print(
        "For information on a command, " "invoke it with the --help flag.",
    )



from hstrat._auxiliary_lib import collapse_nonleading_whitespace


def test_basic_case():
    assert (
        collapse_nonleading_whitespace("This  is    a test")
        == "This is a test"
    )


def test_leading_whitespace():
    assert (
        collapse_nonleading_whitespace("   Leading   whitespace")
        == "   Leading whitespace"
    )


def test_tabs():
    assert (
        collapse_nonleading_whitespace("Tabs\t\tare\there") == "Tabs are here"
    )


def test_mixed_spaces_and_tabs():
    assert (
        collapse_nonleading_whitespace("Mixed \t spaces  and\t\ttabs")
        == "Mixed spaces and tabs"
    )


def test_multiline_string():
    input_text = "Line one    with spaces\nLine two\t\twith tabs"
    expected_output = "Line one with spaces\nLine two with tabs"
    assert collapse_nonleading_whitespace(input_text) == expected_output


def test_no_changes_needed():
    assert (
        collapse_nonleading_whitespace("No extra spaces") == "No extra spaces"
    )


def test_empty_string():
    assert collapse_nonleading_whitespace("") == ""


def test_only_whitespace():
    assert collapse_nonleading_whitespace("     \t \t  ") == "     \t \t  "



from ..... import _auxiliary_lib as auxlib


def calc_provided_uncertainty(
    depth_proportional_resolution: int,
    num_completed_stratum_depositions_since: int,
) -> int:
    """How many ranks forward should the next retained strata be spaced from a
    retained stratum `num_completed_stratum_depositions_since` ranks back in
    the hereditary stratigraphic column's history?

    Calls to this function lay out retained ranks stepwise, with the rank
    chosen from the preceeding call used to determine
    `num_completed_stratum_depositions_since` for the next call. The zeroth
    rank (i.e., most ancient rank) serves as the starting point for the first
    call.

    Note that the returned value actually corresponds to one more than the
    uncertainty with respect to calculating MRCA supposing two identically
    distributed columns. For example a return value of 1 corresponds to
    strata retained at every rank, so the rank of the MRCA can be
    determined with 0 uncertainty.
    """
    guaranteed_resolution = depth_proportional_resolution
    max_uncertainty = (
        num_completed_stratum_depositions_since // guaranteed_resolution
    )

    # round down to lower or equal power of 2
    provided_uncertainty = auxlib.bit_floor(max_uncertainty) or 1
    return provided_uncertainty



import pytest

from hstrat._auxiliary_lib import with_omission


@pytest.mark.parametrize("wrap", [iter, lambda x: x])
def test_with_omission_empty_iterable(wrap):
    iterable = wrap([])
    omit_index = 0
    assert list(with_omission(iterable, omit_index)) == []


@pytest.mark.parametrize("wrap", [iter, lambda x: x])
def test_with_omission_first_index(wrap):
    iterable = wrap([1, 2, 3])
    omit_index = 0
    assert list(with_omission(iterable, omit_index)) == [2, 3]


@pytest.mark.parametrize("wrap", [iter, lambda x: x])
def test_with_omission_middle_index(wrap):
    iterable = wrap([1, 2, 3, 4, 5])
    omit_index = 2
    assert list(with_omission(iterable, omit_index)) == [1, 2, 4, 5]


@pytest.mark.parametrize("wrap", [iter, lambda x: x])
def test_with_omission_last_index(wrap):
    iterable = wrap([1, 2, 3])
    omit_index = 2
    assert list(with_omission(iterable, omit_index)) == [1, 2]


@pytest.mark.parametrize("wrap", [iter, lambda x: x])
def test_with_omission_index_out_of_range(wrap):
    iterable = wrap([1, 2, 3])
    omit_index = 3
    assert list(with_omission(iterable, omit_index)) == [1, 2, 3]


@pytest.mark.parametrize("wrap", [iter, lambda x: x])
def test_with_omission_letters(wrap):
    assert list(with_omission(wrap("abcde"), 2)) == ["a", "b", "d", "e"]



import logging
import typing

import polars as pl
from tqdm import tqdm

from .._auxiliary_lib import (
    alifestd_assign_contiguous_ids,
    alifestd_collapse_unifurcations,
    alifestd_delete_trunk_asexual,
    get_sole_scalar_value_polars,
    log_context_duration,
    log_memory_usage,
    render_pandas_snapshot,
    render_polars_snapshot,
)
from ..phylogenetic_inference.tree.trie_postprocess import NopTriePostprocessor


def surface_postprocess_trie(
    df: pl.DataFrame,
    *,
    trie_postprocessor: typing.Callable = NopTriePostprocessor(),
    # ^^^ NopTriePostprocessor is stateless, so is safe as default value
) -> pl.DataFrame:
    """Postprocess raw phylogenetic tree reconstruction output data to create
    finalized estimate of phylogenetic history.

    Perfoms the following operations:
    - Delete trunk nodes with rank less than `dstream_S`.
    - Collapse unifurcations.
    - Assign contiguous IDs to nodes.
    - Apply supplied `trie_postprocessor` functor.

    Parameters
    ----------
    df : pl.DataFrame
        The input DataFrame containing packed data with required columns, one
        row per genome.

        Required schema:
            - 'id' : pl.UInt64
                - Unique identifier for each taxon (RE alife standard format).
            - 'ancestor_id' : pl.UInt64
                - Unique identifier for ancestor taxon  (RE alife standard
                  format).
            - 'hstrat_rank' : pl.UInt64
                - Num generations elapsed for ancestral differentia.
                - Corresponds to `dstream_Tbar` for inner nodes.
                - Corresponds `dstream_T` - 1 for leaf nodes
            - 'hstrat_differentia_bitwidth' : pl.UInt32
                - Size of annotation differentiae, in bits.
                - Corresponds to `dstream_value_bitwidth`.
            - 'dstream_S' : pl.UInt32
                - Capacity of dstream buffer used for hstrat surface, in number
                  of data items (i.e., differentia values).

        Optional schema:
            - 'dstream_data_id' : pl.UInt64
                - Unique identifier for each genome in source genomedataframe

    trie_postprocessor : Callable, default `hstrat.NopTriePostprocessor()`
        Tree postprocess functor.

        Must take `trie` of type `pandas.DataFrame`,
        `p_differentia_collision` of type `float`, `mutate` of type `bool`, and
        `progress_wrap` of type `Callable` params. Must returned postprocessed
        trie (type `pd.DataFrame`).

        To apply multiple postprocessors, use
        `hstrat.CompoundTriePostprocessor`.

    Returns
    -------
    pl.DataFrame
        The output DataFrame containing the estimated phylogenetic tree in
        alife standard format, with the following columns:

        Required schema:
        - 'id' : pl.UInt64
            - Unique identifier for each taxon (RE alife standard format).
        - 'ancestor_id' : pl.UInt64
            - Unique identifier for ancestor taxon  (RE alife standard format).
        - 'hstrat_rank_from_t0' : pl.UInt64
            - Num generations elapsed for ancestral differentia.
            - Corresponds to `dstream_Tbar` - `dstream_S` for inner nodes.
            - Corresponds `dstream_T` - 1 - `dstream_S` for leaf nodes

        Optional schema:
        - 'origin_tme' : pl.UInt64
            - Estimated origin time for phylogeny nodes, in generations elapsed
              since fouding ancestor.

              Value depends on the trie postprocessor used.

        Additional user-defined columns will be forwarded from the input
        DataFrame. Any columns created by the trie postprocessor will also be
        included.

        Note that the alife-standard `ancestor_list` column is not included in
        the output.

    Notes
    -----
    Collapsing trunk nodes with rank less than `dstream_S` assumes that `S`
    "dummy" strata were added to fill hstrat surface for founding ancestor(s).

    Currently, data is converted to Pandas for processing, then back to Polars.

    See Also
    --------
    surface_unpack_reconstruct :
        Creates raw reconstruction data postprocessed here.
    alifestd_try_add_ancestor_list_col :
        Adds alife-standard `ancestor_list` column to phylogeny data.
    """
    logging.info("beginning surface_postprocess_trie")
    log_memory_usage(logging.info)
    render_polars_snapshot(df, "raw tree", logging.info)
    original_columns = df.columns

    logging.info("extracting differentia bitwidth")
    differentia_bitwidth = get_sole_scalar_value_polars(
        df, "hstrat_differentia_bitwidth"
    )
    logging.info(f" - differentia bitwidth: {differentia_bitwidth}")

    logging.info("converting DataFrame to Pandas...")
    with log_context_duration("pl.DataFrame.to_pandas", logging.info):
        df = df.lazy().collect().to_pandas()
    render_pandas_snapshot(df, "as pandas", logging.info)
    log_memory_usage(logging.info)

    with log_context_duration("alifestd_delete_trunk_asexual", logging.info):
        df["is_trunk"] = df["hstrat_rank"] < df["dstream_S"]
        df = alifestd_delete_trunk_asexual(df, mutate=True)

    with log_context_duration("alifestd_collapse_unifurcations", logging.info):
        df = alifestd_collapse_unifurcations(df, mutate=True)

    render_pandas_snapshot(df, "collapsed tree", logging.info)

    with log_context_duration("alifestd_assign_contiguous_ids", logging.info):
        df = alifestd_assign_contiguous_ids(df, mutate=True)

    render_pandas_snapshot(df, "reassigned tree", logging.info)

    with log_context_duration("trie_postprocessor", logging.info):
        pre_postprocessor_columns = {*df.columns}
        df = df.rename(columns={"hstrat_rank": "rank"})
        df = trie_postprocessor(
            df,
            p_differentia_collision=2**-differentia_bitwidth,
            mutate=True,
            progress_wrap=tqdm,
        )
        df = df.rename(columns={"rank": "hstrat_rank"})

    render_pandas_snapshot(df, "with trie postprocessing", logging.info)

    logging.info("setting up hstrat_rank_from_t0...")
    df["hstrat_rank_from_t0"] = df["hstrat_rank"] - df["dstream_S"]

    to_keep = {*original_columns} - {
        "dstream_S",
        "hstrat_differentia_bitwidth",
        "hstrat_rank",
    }
    to_drop = pre_postprocessor_columns - to_keep
    logging.info(f"dropping columns {to_drop=}...")
    df.drop(columns=[*to_drop], inplace=True)

    logging.info("converting DataFrame to Polars...")
    with log_context_duration("pl.from_pandas", logging.info):
        df = pl.from_pandas(df)
    render_polars_snapshot(df, "as polars", logging.info)
    log_memory_usage(logging.info)

    return df



import random
import typing

import pandas as pd

from ._alifestd_find_leaf_ids import alifestd_find_leaf_ids
from ._alifestd_prune_extinct_lineages_asexual import (
    alifestd_prune_extinct_lineages_asexual,
)
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col
from ._with_rng_state_context import with_rng_state_context


def _alifestd_downsample_tips_asexual_impl(
    phylogeny_df: pd.DataFrame,
    n_downsample: int,
) -> pd.DataFrame:
    """Implementation detail for alifestd_downsample_tips_asexual."""
    tips = alifestd_find_leaf_ids(phylogeny_df)
    kept = random.sample(tips, min(n_downsample, len(tips)))
    phylogeny_df["extant"] = phylogeny_df["id"].isin(kept)

    return alifestd_prune_extinct_lineages_asexual(
        phylogeny_df, mutate=True
    ).drop(columns=["extant"])


def alifestd_downsample_tips_asexual(
    phylogeny_df: pd.DataFrame,
    n_downsample: int,
    mutate: bool = False,
    seed: typing.Optional[int] = None,
) -> pd.DataFrame:
    """Subsample phylogeny containing `num_tips` tips. If `num_tips` is greater
    than the number of tips in the phylogeny, the whole phylogeny is returned.

    Only supports asexual phylogenies.
    """
    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df)
    if "ancestor_id" not in phylogeny_df.columns:
        raise ValueError(
            "alifestd_downsample_tips_asexual only supports "
            "asexual phylogenies.",
        )

    if phylogeny_df.empty:
        return phylogeny_df

    impl = (
        with_rng_state_context(seed)(_alifestd_downsample_tips_asexual_impl)
        if seed is not None
        else _alifestd_downsample_tips_asexual_impl
    )

    return impl(phylogeny_df, n_downsample)



import itertools as it
import os

import alifedata_phyloinformatics_convert as apc
import pandas as pd
import pytest

from hstrat import hstrat
from hstrat._auxiliary_lib import seed_random

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "stratum_ordered_store",
    hstrat.provided_stratum_ordered_stores,
)
@pytest.mark.parametrize(
    "stratum_retention_algorithm",
    hstrat.provided_stratum_retention_algorithms,
)
def test_genome_instrumentation_deterministic(
    stratum_ordered_store, stratum_retention_algorithm
):
    try:
        policy = stratum_retention_algorithm.Policy()
    except TypeError:
        policy = stratum_retention_algorithm.Policy(1)

    derived_columns = []
    for _rep in range(3):
        seed_random(1)
        col = hstrat.HereditaryStratigraphicColumn(
            stratum_retention_policy=policy,
            stratum_ordered_store=stratum_ordered_store,
        )
        for __ in range(50):
            col.DepositStratum()

        col.DepositStrata(10)

        col = col.CloneDescendant()

        col = col.CloneNthDescendant(10)

        derived_columns.append(col)

    for a, b in it.combinations(derived_columns, 2):
        assert a == b

    # control test
    derived_columns = []
    for rep in range(3):
        seed_random(rep)
        col = hstrat.HereditaryStratigraphicColumn(
            stratum_retention_policy=policy,
            stratum_ordered_store=stratum_ordered_store,
        )
        for __ in range(50):
            col.DepositStratum()

        col.DepositStrata(10)

        col = col.CloneDescendant()

        col = col.CloneNthDescendant(10)

        derived_columns.append(col)

    for a, b in it.combinations(derived_columns, 2):
        assert a != b


def test_template_phylogeny_generation_determinstic():
    derived_phylogenetic_records = []
    for _rep in range(3):
        seed_random(1)

        derived_phylogenetic_records.append(
            hstrat.evolve_fitness_trait_population()
        )

    for a, b in it.combinations(derived_phylogenetic_records, 2):
        assert a.equals(b)

    # control test
    derived_phylogenetic_records = []
    for rep in range(3):
        seed_random(rep)

        derived_phylogenetic_records.append(
            hstrat.evolve_fitness_trait_population()
        )

    for a, b in it.combinations(derived_phylogenetic_records, 2):
        assert not a.equals(b)


@pytest.mark.parametrize(
    "stratum_retention_algorithm",
    hstrat.provided_stratum_retention_algorithms,
)
@pytest.mark.parametrize(
    "tree",
    [
        apc.alife_dataframe_to_dendropy_tree(
            pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
            setup_edge_lengths=True,
        ),
        apc.alife_dataframe_to_dendropy_tree(
            pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
            setup_edge_lengths=True,
        ),
        apc.alife_dataframe_to_dendropy_tree(
            pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
            setup_edge_lengths=True,
        ),
    ],
)
def test_template_phylogeny_descent_deterministic(
    stratum_retention_algorithm,
    tree,
):
    try:
        policy = stratum_retention_algorithm.Policy()
    except TypeError:
        policy = stratum_retention_algorithm.Policy(1)

    derived_column_populations = []
    for _rep in range(3):
        seed_random(1)

        seed_column = hstrat.HereditaryStratigraphicColumn(
            stratum_retention_policy=policy,
        )

        extant_population = hstrat.descend_template_phylogeny(
            ascending_lineage_iterables=(
                tip_node.ancestor_iter(
                    inclusive=True,
                )
                for tip_node in tree.leaf_node_iter()
            ),
            descending_tree_iterable=tree.levelorder_node_iter(),
            get_parent=lambda node: node.parent_node,
            get_stem_length=lambda node: node.edge_length,
            seed_column=seed_column,
        )

        derived_column_populations.append(extant_population)

    for a, b in it.combinations(derived_column_populations, 2):
        assert a == b

    # control test
    derived_column_populations = []
    for rep in range(3):
        seed_random(rep)

        seed_column = hstrat.HereditaryStratigraphicColumn(
            stratum_retention_policy=policy,
        )

        extant_population = hstrat.descend_template_phylogeny(
            ascending_lineage_iterables=(
                tip_node.ancestor_iter(
                    inclusive=True,
                )
                for tip_node in tree.leaf_node_iter()
            ),
            descending_tree_iterable=tree.levelorder_node_iter(),
            get_parent=lambda node: node.parent_node,
            get_stem_length=lambda node: node.edge_length,
            seed_column=seed_column,
        )

        derived_column_populations.append(extant_population)

    for a, b in it.combinations(derived_column_populations, 2):
        assert a != b



import functools


@functools.lru_cache()
def calc_provided_degree(
    size_curb: int,
    interspersal: int,
    num_stratum_depositions_completed: int,
) -> int:
    """After m strata have been deposited, what geometric sequence nth root
    algorithm degree n is provided?"""
    # size_limit_inclusive = degree * 2 * (interspersal + 1) + 2
    # degree = (size_limit_inclusive - 2) / (2 * interspersal + 2)
    res = max(
        (size_curb - 2) // (2 * interspersal + 2),
        1,
    )

    return res



import typing

import anytree

from ._AnyTreeFastPostOrderIter import AnyTreeFastPostOrderIter


def anytree_calc_leaf_counts(tree: anytree.Node) -> typing.Dict[int, int]:
    """Prepare a dict that maps each node id in a tree to the leaf count of its
    subtree.

    Parameters
    ----------
    node : anytree.Node
        The root node of the Anytree tree to compute the size of.
    """

    res = dict()
    for node in AnyTreeFastPostOrderIter(tree):
        if node.is_leaf:
            res[id(node)] = 1
        else:
            res[id(node)] = sum(res[id(child)] for child in node.children)
    return res



import math
import typing

from ..._auxiliary_lib import HereditaryStratigraphicArtifact
from ..priors import ArbitraryPrior, UniformPrior
from ..priors._BubbleWrappedPrior import BubbleWrappedPrior
from ..priors._detail import PriorBase
from ._impl import (
    estimate_rank_of_mrca_maximum_likelihood,
    estimate_rank_of_mrca_naive,
    estimate_rank_of_mrca_unbiased,
)


def estimate_rank_of_mrca_between(
    first: HereditaryStratigraphicArtifact,
    second: HereditaryStratigraphicArtifact,
    estimator: str,
    prior: typing.Union[typing.Literal["arbitrary", "uniform"], PriorBase],
) -> typing.Optional[float]:
    """At what generation did the most recent common ancestor of first and
    second occur?

    Parameters
    ----------
    estimator : {"maximum_likelihood", "unbiased"}
        What estimation method should be used? Options are "maximum_likelihood"
        or "unbiased".

        The "maximum_likelihood" estimator is faster to compute than the
        "unbiased" estimator.
    prior : {"arbitrary", "uniform"} or object implementing prior interface
        Prior probability density distribution over possible generations of the
        MRCA.

        Implementations for arbitrary, geometric, exponential, and uniform
        priors are available in hstrat.phylogenetic_inference.priors. User
        -defined classes specifying custom priors can also be provided.

    Returns
    -------
    float, optional
        Estimate of MRCA rank, unless first and second definitively share no
        common ancestor in which case None will be returned.

    See Also
    --------
    calc_rank_of_mrca_bounds_between :
        Calculates confidence intervals for generation of most recent common
        ancestor between two hereditary stratigraphic columns.
    does_definitively_have_no_common_anestor :
        Does the hereditary stratigraphic record definitively prove that first
        and second could not possibly share a common ancestor?
    """

    if estimator == "maximium_likelihood" and prior == "arbitrary":
        return estimate_rank_of_mrca_naive(first, second)

    estimator = {
        "maximum_likelihood": estimate_rank_of_mrca_maximum_likelihood,
        "unbiased": estimate_rank_of_mrca_unbiased,
    }[estimator]

    if isinstance(prior, str):
        prior = {
            "arbitrary": ArbitraryPrior,
            "uniform": UniformPrior,
        }[prior]()

    res = estimator(
        first,
        second,
        prior=BubbleWrappedPrior(prior),
    )
    if res is not None:
        assert 0 <= res or math.isclose(0, res, abs_tol=10e-6), res
        max_ = (
            min(first.GetNumStrataDeposited(), second.GetNumStrataDeposited())
            - 1
        )
        assert res <= max_ or math.isclose(res, max_), (res, max_)
    return res



#!/usr/bin/env python3

from hstrat import hstrat

if __name__ == "__main__":

    print("creating founder1 and founder2, which share no ancestry...")
    founder1 = hstrat.HereditaryStratigraphicColumn(
        # retain strata from every generation
        stratum_retention_policy=hstrat.fixed_resolution_algo.Policy(1)
    )
    founder2 = hstrat.HereditaryStratigraphicColumn(
        # retain strata from every third generation
        stratum_retention_policy=hstrat.fixed_resolution_algo.Policy(3),
    )

    print(
        "   ",
        "do founder1 and founder2 share any common ancestor?",
        hstrat.does_have_any_common_ancestor(founder1, founder2),
    )

    print()
    print("creating child1 from founder1...")
    child1 = founder1.CloneDescendant()

    print(
        "   ",
        "do founder1 and child1 share any common ancestor?",
        hstrat.does_have_any_common_ancestor(founder1, child1),
    )

    print(
        "   ",
        "do founder2 and child1 share any common ancestor?",
        hstrat.does_have_any_common_ancestor(founder2, child1),
    )

    print()
    print("creating descendant1a, 10 generations removed from founder1...")
    descendant1a = founder1.Clone()
    for __ in range(10):
        descendant1a.DepositStratum()

    print("creating descendant1b, 50 generations removed from founder1...")
    descendant1b = founder1.Clone()
    for __ in range(50):
        descendant1b.DepositStratum()

    print(
        "   ",
        "estimate descendant1a generations since MRCA with descendant1b?",
        hstrat.calc_ranks_since_mrca_bounds_with(
            descendant1a, descendant1b, prior="arbitrary"
        ),
    )
    print(
        "   ",
        "estimate descendant1b generations since MRCA with descendant1a?",
        hstrat.calc_ranks_since_mrca_bounds_with(
            descendant1b, descendant1a, prior="arbitrary"
        ),
    )
    print(
        "estimate generation of MRCA between descendant1a and descendant1b?",
        hstrat.calc_rank_of_mrca_bounds_between(
            descendant1a, descendant1b, prior="arbitrary"
        ),
    )

    print()
    print("creating descendant2a, 10 generations removed from founder2...")
    descendant2a = founder2.Clone()
    for __ in range(10):
        descendant2a.DepositStratum()
    print("creating descendant2b, 20 generations removed from descendant2a...")
    descendant2b = descendant2a.Clone()
    for __ in range(20):
        descendant2b.DepositStratum()
    print("creating descendant2c, 5 generations removed from descendant2a...")
    descendant2c = descendant2a.Clone()
    for __ in range(5):
        descendant2c.DepositStratum()

    print(
        "   ",
        "estimate descendant2b generations since MRCA with descendant2c?",
        hstrat.calc_ranks_since_mrca_bounds_with(
            descendant2b, descendant2c, prior="arbitrary"
        ),
    )
    print(
        "   ",
        "estimate descendant2c generations since MRCA with descendant2b?",
        hstrat.calc_ranks_since_mrca_bounds_with(
            descendant2c, descendant2b, prior="arbitrary"
        ),
    )
    print(
        "   ",
        "estimate generation of MRCA between descendant2b and descendant2c?",
        hstrat.calc_rank_of_mrca_bounds_between(
            descendant2b, descendant2c, prior="arbitrary"
        ),
    )
    print(
        "    ^^^ "
        "note MRCA estimate uncertainty, "
        "caused by sparser stratum retention policy"
    )



import dataclasses
import itertools as it
import typing

import opytional as opyt

from ._SearchtableRecord import SearchtableRecord


class Searchtable:
    """Tabular data structure for representing phylogeny reconstruction trie,
    with reconfigurable "search trie" pointers to collapse portions of the trie
    for efficient search.

    Although "search trie" pointers are used to collapse portions of the trie,
    "build trie" pointers are used to record the original structure of the trie.

    Implementation detail for `build_tree_searchtable_python`.
    """

    _records: typing.List[SearchtableRecord]

    def __init__(self: "Searchtable") -> None:
        """Initialize the search table with a root node record."""
        self._records = [
            SearchtableRecord(rank=0, taxon_id=0, taxon_label="_root"),
        ]

    def get_differentia_of(self: "Searchtable", taxon_id: int) -> int:
        """Returns the differentia of the given taxon."""
        return self._records[taxon_id].differentia

    def get_rank_of(self: "Searchtable", taxon_id: int) -> int:
        """Returns the rank of the given taxon."""
        return self._records[taxon_id].rank

    def to_records(self: "Searchtable") -> typing.List[dict]:
        """Returns list of dictionaries corresponding to "build trie" nodes."""
        return [*map(dataclasses.asdict, self._records)]

    def has_search_parent(self: "Searchtable", taxon_id: int) -> bool:
        """Does the given node have parent in the search trie?"""
        return self._records[taxon_id].search_ancestor_id != taxon_id

    def iter_search_children_of(
        self: "Searchtable", taxon_id: int
    ) -> typing.Iterable[int]:
        """Iterate over the children of the given node, with respect to
        "search trie" structure."""
        prev = taxon_id
        cur = self._records[taxon_id].search_first_child_id
        while cur != prev:
            yield cur
            prev = cur
            cur = self._records[cur].search_next_sibling_id

    def iter_inner_search_children_of(
        self: "Searchtable",
        taxon_id: int,
    ) -> typing.Iterable[int]:
        """Filter `iter_search_children_of` to exclude leaf nodes."""
        for child in self.iter_search_children_of(taxon_id):
            if self._records[child].search_first_child_id != child:
                yield child

    def attach_search_parent(
        self: "Searchtable", *, taxon_id: int, parent_id: int
    ) -> None:
        """Attach a parent to the given node in the search trie, updating
        search trie pointers accordingly."""
        if self._records[taxon_id].search_ancestor_id == parent_id:
            return

        self._records[taxon_id].search_ancestor_id = parent_id

        ancestor_first_child = self._records[parent_id].search_first_child_id
        is_first_child_ = ancestor_first_child == parent_id
        new_next_sibling = (
            taxon_id if is_first_child_ else ancestor_first_child
        )
        self._records[taxon_id].search_next_sibling_id = new_next_sibling
        self._records[parent_id].search_first_child_id = taxon_id

    def detach_search_parent(self: "Searchtable", taxon_id: int) -> None:
        """Detach the parent of the given node in the search trie, leaving
        `taxon_id` with no search parent."""
        ancestor_id = self._records[taxon_id].search_ancestor_id
        assert self.has_search_parent(taxon_id)

        is_first_child_ = (
            self._records[ancestor_id].search_first_child_id == taxon_id
        )
        next_sibling = self._records[taxon_id].search_next_sibling_id
        is_last_child = next_sibling == taxon_id

        if is_first_child_:
            new_first_child = ancestor_id if is_last_child else next_sibling
            self._records[ancestor_id].search_first_child_id = new_first_child
        else:
            for child1, child2 in it.pairwise(
                self.iter_search_children_of(ancestor_id),
            ):
                if child2 == taxon_id:
                    new_next_sib = child1 if is_last_child else next_sibling
                    self._records[child1].search_next_sibling_id = new_next_sib
                    break
            else:
                assert False

        self._records[taxon_id].search_ancestor_id = taxon_id
        self._records[taxon_id].search_next_sibling_id = taxon_id

    def create_offspring(
        self: "Searchtable",
        *,
        parent_id: int,
        differentia: int,
        rank: int,
        taxon_label: typing.Optional[str] = None,
    ) -> int:
        """Create a new node in the search trie, returning its taxon ID."""

        taxon_id = len(self._records)
        record = SearchtableRecord(
            ancestor_id=parent_id,
            differentia=differentia,
            rank=rank,
            taxon_id=taxon_id,
            taxon_label=opyt.or_value(taxon_label, f"{len(self._records)}"),
            search_ancestor_id=taxon_id,  # will be set in attach_search_parent
            search_first_child_id=taxon_id,
            search_next_sibling_id=taxon_id,
        )
        self._records.append(record)

        assert not self.has_search_parent(taxon_id)
        self.attach_search_parent(taxon_id=taxon_id, parent_id=parent_id)

        return taxon_id



import typing
import warnings

from .._detail import PolicySpecBase


class PolicySpec(PolicySpecBase):
    """Contains all policy parameters, if any."""

    _degree: int
    _interspersal: int

    def __init__(
        self: "PolicySpec",
        degree: int,
        interspersal: int = 2,
    ) -> None:
        """Construct the policy spec.

        Parameters
        ----------
        degree : int
            How many should target recencies for uncertainty-capped coverage
            should be spaced exponentially from zero recency to maximum recency
            (i.e., number strata deposited)? Adjust this parameter to set upper
            bound on space complexity (i.e., to ensure available space is not
            exceeded).
        interspersal : int
            At least how many retained ranks should be spaced between zero
            recency and each target recency? Must be greater than 0. No bound
            on MRCA rank estimate uncertainty provided if set to 1. For most
            use cases, leave this set to 2.
        """
        assert degree >= 0
        assert interspersal >= 1
        if interspersal == 1:
            warnings.warn(
                "Interspersal set to 1, "
                "no bound on MRCA rank estimate uncertainty can be guaranteed.",
            )
        self._degree = degree
        self._interspersal = interspersal

    def __eq__(self: "PolicySpec", other: typing.Any) -> bool:
        return isinstance(other, self.__class__) and (
            self._degree,
            self._interspersal,
        ) == (
            other._degree,
            other._interspersal,
        )

    def __hash__(self: "PolicySpec") -> int:
        """Hash object instance."""
        return hash(
            (
                self._degree,
                self._interspersal,
            )
        )

    def __repr__(self: "PolicySpec") -> str:
        return f"""{
            self.GetAlgoIdentifier()
        }.{
            PolicySpec.__qualname__
        }(degree={
            self._degree
        }, interspersal={
            self._interspersal
        })"""

    def __str__(self: "PolicySpec") -> str:
        return f"""{
            self.GetAlgoTitle()
        } (degree: {
            self._degree
        }, interspersal: {
            self._interspersal
        })"""

    def GetEvalCtor(self: "PolicySpec") -> str:
        return f"hstrat.{self!r}"

    def GetDegree(self: "PolicySpec") -> int:
        return self._degree

    def GetInterspersal(self: "PolicySpec") -> int:
        return self._interspersal

    @staticmethod
    def GetAlgoIdentifier() -> str:
        """Get programatic name for underlying retention algorithm."""
        return __package__.split(".")[-1]

    @staticmethod
    def GetAlgoTitle() -> str:
        """Get human-readable name for underlying retention algorithm."""
        return (
            "Tapered Nth Root Geometric Sequence Stratum Retention Algorithm"
        )



"""Accept linear size complexity to maintain fixed spacing between strata.

The fixed resolution policy ensures estimates of MRCA rank will have
uncertainty bounds less than or equal a fixed, absolute user-specified cap
that is independent of the number of strata deposited on either column.
Thus, MRCA rank estimate uncertainty scales as O(1) with respect to the
greater number of strata deposited on either column.

Under the fixed resolution policy, the number of strata retained (i.e.,
space complexity) scales as O(n) with respect to the number of strata
deposited.
"""

from ...._auxiliary_lib import lazy_attach_stub

__getattr__, __dir__, __all__ = lazy_attach_stub(__name__, __file__)
del lazy_attach_stub



import typing

from ..._detail import PolicyCouplerBase
from .._PolicySpec import PolicySpec
from .._impl import calc_provided_uncertainty


class CalcRankAtColumnIndex:
    """Functor to provide member function implementation in Policy class."""

    def __init__(
        self: "CalcRankAtColumnIndex",
        policy_spec: typing.Optional[PolicySpec],
    ) -> None:
        pass

    def __eq__(self: "CalcRankAtColumnIndex", other: typing.Any) -> bool:
        return isinstance(other, self.__class__)

    def _CalcRankAtColumnIndexImpl(
        self: "CalcRankAtColumnIndex",
        policy: PolicyCouplerBase,
        index: int,
        num_strata_deposited: int,
    ) -> int:
        """After n strata have been deposited, what will the rank of the
        stratum at column index k be?

        Assumes that no in-progress stratum depositions that haven't been
        reflected in num_strata_deposited.
        """
        spec = policy.GetSpec()
        guaranteed_resolution = spec.GetDepthProportionalResolution()

        cur_stage_uncertainty = calc_provided_uncertainty(
            guaranteed_resolution,
            num_strata_deposited,
        )

        prev_stage_uncertainty = cur_stage_uncertainty // 2
        prev_stage_max_idx = (
            num_strata_deposited - 2
        ) // prev_stage_uncertainty

        thresh_idx = (
            2 * prev_stage_max_idx - 4 * guaranteed_resolution + 2
        ) // 2

        before_thresh_idx = min(thresh_idx, index)
        after_thresh_idx = max(index - thresh_idx, 0)

        return (
            before_thresh_idx * cur_stage_uncertainty
            + after_thresh_idx * prev_stage_uncertainty
        )

    def __call__(
        self: "CalcRankAtColumnIndex",
        policy: typing.Optional[PolicyCouplerBase],
        index: int,
        num_strata_deposited: typing.Optional[int],
    ) -> int:
        """After n strata have been deposited, what will the rank of the
        stratum at column index k be?

        Enables a HereditaryStratigraphicColumn using this predicate to
        optimize away storage of rank annotations on strata. Takes into the
        account the possibility for in-progress stratum depositions that haven't
        been reflected in num_strata_deposited.
        """
        spec = policy.GetSpec()
        guaranteed_resolution = spec.GetDepthProportionalResolution()

        if num_strata_deposited < guaranteed_resolution * 2 + 1:
            # use identity mapping before first ranks are condemned
            return index
        elif (
            index
            == policy.CalcNumStrataRetainedExact(num_strata_deposited) - 1
        ):
            # case where index is the very most recent stratum
            return num_strata_deposited - 1
        elif index == policy.CalcNumStrataRetainedExact(num_strata_deposited):
            # in cases where the index is an in-progress
            # deposition rank must be calculated as the rank succeeding the
            # previous stratum's rank
            return (
                self(
                    policy,
                    index - 1,
                    num_strata_deposited,
                )
                + 1
            )
        else:
            return self._CalcRankAtColumnIndexImpl(
                policy,
                index,
                num_strata_deposited,
            )



import pytest

from hstrat._auxiliary_lib import bit_drop_msb


@pytest.mark.parametrize(
    "input_, expected_output",
    [
        (0, 0),
        (1, 0),
        (2, 0),
        (3, 1),
        (4, 0),
        (5, 1),
        (6, 2),
        (7, 3),
        (8, 0),
        (9, 1),
        (10, 2),
        (255, 127),
        (256, 0),
        (257, 1),
        (1234567890, 160826066),
    ],
)
def test_bit_drop_msb(input_, expected_output):
    assert bit_drop_msb(input_) == expected_output



import pandas as pd

from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col


def alifestd_count_children_of_asexual(
    phylogeny_df: pd.DataFrame,
    parent: int,
    mutate: bool = False,
) -> int:
    """How many taxa are direct descendants of the given parent?

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)

    if not (phylogeny_df["id"] == parent).any():
        raise ValueError(f"Parent {parent} not found in phylogeny dataframe.")

    is_root = phylogeny_df["ancestor_id"] == phylogeny_df["id"]
    return ((phylogeny_df["ancestor_id"] == parent) & ~is_root).sum()



def calc_geom_seq_nth_root_transition_rank(size_curb: int) -> int:
    """What is the first rank where recency-proportional resolution algorithm
    sunsets and geometric sequence nth root algorithm applies?

    Return value zero implies that recency-proportional resolution algorithm
    never applies and geometric sequence nth root algorithm is used throughout
    deposition sequence.
    """
    # resolution_thresh threshold chosen to ensure availability of ranks
    # required by geometric sequence nth root algorithm at transion point
    # see calc_provided_resolution, where equivalent threshold also applies
    resolution_thresh = 2
    assert resolution_thresh >= 0
    return 2 ** (size_curb // (resolution_thresh + 1)) // 2



import itertools as it
import typing

import lru


def memoize_generator(lru_cache_size: int = 128) -> typing.Callable:
    """Decorator to memoize a generator function using an LRU cache.

    The cache stores the original generator and a list of its values.
    When the generator is called, if the arguments are in the cache,
    the cached values are returned, otherwise the generator is executed and its values are stored.

    Parameters
    ----------
    lru_cache_size : int, default 128
        The maximum size of the LRU cache.

    Returns
    -------
    callable
        Decorator function to memoize a generator.

    Examples
    --------
    >>> @memoize_generator(lru_cache_size=2)
    ... def gen(n):
    ...     for i in range(n):
    ...         yield i
    ...
    >>> g = gen(3)
    >>> list(g)
    [0, 1, 2]
    >>> g = gen(3)
    >>> list(g)
    [0, 1, 2]
    """
    # adapted from https://stackoverflow.com/a/10726355/17332200
    def decorator(generator: typing.Callable) -> typing.Callable:
        cache_original = lru.LRU(lru_cache_size)
        cache_copy = lru.LRU(lru_cache_size)

        def ret(*args) -> typing.Generator:
            if args not in cache_original:
                cache_original[args] = generator(*args)
                cache_copy[args] = []

            original = cache_original[args]
            copy = cache_copy[args]

            for idx in it.count():
                if idx < len(copy):
                    yield copy[idx]
                else:
                    assert idx == len(copy)
                    try:
                        copy.append(next(original))
                        yield copy[idx]
                    except StopIteration:
                        return

        return ret

    return decorator



from animate import policy_panel_animate, stratum_retention_animate
from ascii import col_to_ascii
from plot import (
    mrca_uncertainty_absolute_barplot,
    mrca_uncertainty_relative_barplot,
    policy_panel_plot,
    strata_retained_frac_lineplot,
    strata_retained_num_lineplot,
    stratum_retention_dripplot,
)

from . import animate, ascii, plot

__all__ = [
    "animate",
    "ascii",
    "plot",
    "col_to_ascii",
    "stratum_retention_animate",
    "policy_panel_animate",
    "mrca_uncertainty_absolute_barplot",
    "mrca_uncertainty_relative_barplot",
    "policy_panel_plot",
    "strata_retained_frac_lineplot",
    "strata_retained_num_lineplot",
    "stratum_retention_dripplot",
]



import pandas as pd


def alifestd_has_compact_ids(phylogeny_df: pd.DataFrame) -> bool:
    """Are id values between 0 and `len(phylogeny_df)`, in any order?

    Input dataframe is not mutated by this operation.
    """
    return len(phylogeny_df) == 0 or (
        phylogeny_df["id"].max() == len(phylogeny_df) - 1
    )



import random

import numpy as np
import pytest

from hstrat import hstrat


@pytest.mark.parametrize(
    "differentia_bit_width",
    [1, 2, 8, 64],
)
@pytest.mark.parametrize(
    "estimator",
    ["maximum_likelihood", "unbiased"],
)
@pytest.mark.parametrize(
    "prior",
    ["arbitrary", "uniform"],
)
@pytest.mark.parametrize(
    "force_common_ancestry",
    [True, False, None],
)
def test_build_distance_matrix_numpy_empty(
    differentia_bit_width,
    estimator,
    prior,
    force_common_ancestry,
):

    population = []
    assert np.array_equal(
        hstrat.build_distance_matrix_numpy(
            population,
            estimator,
            prior,
            force_common_ancestry,
        ),
        np.zeros((0, 0)),
    )


@pytest.mark.parametrize(
    "differentia_bit_width",
    [1, 2, 8, 64],
)
@pytest.mark.parametrize(
    "estimator",
    ["maximum_likelihood", "unbiased"],
)
@pytest.mark.parametrize(
    "prior",
    ["arbitrary", "uniform"],
)
@pytest.mark.parametrize(
    "force_common_ancestry",
    [True, False, None],
)
@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_build_distance_matrix_numpy_singleton(
    differentia_bit_width,
    estimator,
    prior,
    force_common_ancestry,
    wrap,
):

    population = [wrap(hstrat.HereditaryStratigraphicColumn())]
    assert np.array_equal(
        hstrat.build_distance_matrix_numpy(
            population,
            estimator,
            prior,
            force_common_ancestry,
        ),
        np.array([[0.0]]),
    )


@pytest.mark.filterwarnings(
    "ignore:Insufficient common ranks between columns to detect common ancestry at given confidence level."
)
@pytest.mark.parametrize(
    "differentia_bit_width",
    [64],
)
@pytest.mark.parametrize(
    "estimator",
    [
        "maximum_likelihood",
        "unbiased",
    ],
)
@pytest.mark.parametrize(
    "prior",
    ["arbitrary", "uniform"],
)
@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_build_distance_matrix_numpy_pair_unrelated(
    differentia_bit_width,
    estimator,
    prior,
    wrap,
):

    population = [
        wrap(hstrat.HereditaryStratigraphicColumn()),
        wrap(hstrat.HereditaryStratigraphicColumn()),
    ]
    assert np.array_equal(
        hstrat.build_distance_matrix_numpy(
            population,
            estimator,
            prior,
            force_common_ancestry=True,
        ),
        np.array([[0.0, 2.0], [2.0, 0.0]]),
    )
    assert np.array_equal(
        hstrat.build_distance_matrix_numpy(
            population,
            estimator,
            prior,
            force_common_ancestry=False,
        ),
        np.array([[0.0, np.nan], [np.nan, 0.0]]),
        equal_nan=True,
    )
    with pytest.raises(ValueError):
        hstrat.build_distance_matrix_numpy(
            population,
            estimator,
            prior,
            force_common_ancestry=None,
        )


@pytest.mark.parametrize(
    "generations_before_mrca",
    [0, 8],
)
@pytest.mark.parametrize(
    "first_generations_since_mrca",
    [0, 5],
)
@pytest.mark.parametrize(
    "second_generations_since_mrca",
    [0, 5],
)
@pytest.mark.parametrize(
    "differentia_bit_width",
    [1, 8, 64],
)
@pytest.mark.parametrize(
    "estimator",
    [
        "maximum_likelihood",
        "unbiased",
    ],
)
@pytest.mark.parametrize(
    "prior",
    ["arbitrary", "uniform", hstrat.GeometricPrior(1.1)],
)
@pytest.mark.parametrize(
    "force_common_ancestry",
    [True, False, None],
)
@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_build_distance_matrix_numpy_pair_related(
    generations_before_mrca,
    first_generations_since_mrca,
    second_generations_since_mrca,
    differentia_bit_width,
    estimator,
    prior,
    force_common_ancestry,
    wrap,
):

    if differentia_bit_width == 64:
        common_ancestor = hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=differentia_bit_width
        ).CloneNthDescendant(generations_before_mrca)
        population = [
            wrap(
                common_ancestor.CloneNthDescendant(
                    first_generations_since_mrca
                )
            ),
            wrap(
                common_ancestor.CloneNthDescendant(
                    second_generations_since_mrca
                )
            ),
        ]
        expected_patristic_distance = (
            first_generations_since_mrca + second_generations_since_mrca
        )
        assert np.isclose(
            hstrat.build_distance_matrix_numpy(
                population,
                estimator,
                prior,
                force_common_ancestry=force_common_ancestry,
            ),
            np.array(
                [
                    [0.0, expected_patristic_distance],
                    [expected_patristic_distance, 0.0],
                ]
            ),
        ).all()

    common_ancestor = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=hstrat.fixed_resolution_algo.Policy(3),
        stratum_differentia_bit_width=differentia_bit_width,
    ).CloneNthDescendant(generations_before_mrca)
    population = [
        wrap(common_ancestor.CloneNthDescendant(first_generations_since_mrca)),
        wrap(
            common_ancestor.CloneNthDescendant(second_generations_since_mrca)
        ),
    ]
    expected_patristic_distance = hstrat.estimate_patristic_distance_between(
        *population,
        estimator=estimator,
        prior=prior,
    )
    assert np.isclose(
        hstrat.build_distance_matrix_numpy(
            population,
            estimator=estimator,
            prior=prior,
            force_common_ancestry=force_common_ancestry,
        ),
        np.array(
            [
                [0.0, expected_patristic_distance],
                [expected_patristic_distance, 0.0],
            ]
        ),
    ).all()


@pytest.mark.parametrize(
    "generations_before_mrca",
    [0, 8],
)
@pytest.mark.parametrize(
    "first_generations_since_mrca",
    [0, 5],
)
@pytest.mark.parametrize(
    "second_generations_since_mrca",
    [0, 5],
)
@pytest.mark.parametrize(
    "differentia_bit_width",
    [64],
)
@pytest.mark.parametrize(
    "estimator",
    [
        "maximum_likelihood",
        "unbiased",
    ],
)
@pytest.mark.parametrize(
    "prior",
    ["arbitrary", "uniform", hstrat.GeometricPrior(1.1)],
)
@pytest.mark.parametrize(
    "force_common_ancestry",
    [True, False, None],
)
@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_build_distance_matrix_numpy_pair_and_unrelated(
    generations_before_mrca,
    first_generations_since_mrca,
    second_generations_since_mrca,
    differentia_bit_width,
    estimator,
    prior,
    force_common_ancestry,
    wrap,
):

    common_ancestor = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=differentia_bit_width
    ).CloneNthDescendant(generations_before_mrca)
    population = [
        wrap(common_ancestor.CloneNthDescendant(first_generations_since_mrca)),
        wrap(
            common_ancestor.CloneNthDescendant(second_generations_since_mrca)
        ),
        wrap(
            hstrat.HereditaryStratigraphicColumn(
                stratum_differentia_bit_width=differentia_bit_width
            )
        ),
    ]
    expected_patristic_distance = (
        first_generations_since_mrca + second_generations_since_mrca
    )
    assert np.array_equal(
        hstrat.build_distance_matrix_numpy(
            population,
            estimator,
            prior,
            force_common_ancestry=False,
        ),
        np.array(
            [
                [0.0, expected_patristic_distance, np.nan],
                [expected_patristic_distance, 0.0, np.nan],
                [np.nan, np.nan, 0.0],
            ]
        ),
        equal_nan=True,
    )

    max_patristic_distance0to2 = (
        population[0].GetNumStrataDeposited()
        + population[2].GetNumStrataDeposited()
    )
    max_patristic_distance1to2 = (
        population[1].GetNumStrataDeposited()
        + population[2].GetNumStrataDeposited()
    )
    assert np.array_equal(
        hstrat.build_distance_matrix_numpy(
            population,
            estimator,
            prior,
            force_common_ancestry=True,
        ),
        np.array(
            [
                [0.0, expected_patristic_distance, max_patristic_distance0to2],
                [expected_patristic_distance, 0.0, max_patristic_distance1to2],
                [max_patristic_distance0to2, max_patristic_distance1to2, 0.0],
            ]
        ),
    )


@pytest.mark.parametrize(
    "generations_before_mrca",
    [0, 8],
)
@pytest.mark.parametrize(
    "differentia_bit_width",
    [1, 8, 64],
)
@pytest.mark.parametrize(
    "estimator",
    [
        "maximum_likelihood",
        "unbiased",
    ],
)
@pytest.mark.parametrize(
    "prior",
    ["arbitrary", "uniform", hstrat.GeometricPrior(1.1)],
)
@pytest.mark.parametrize(
    "force_common_ancestry",
    [True, False, None],
)
@pytest.mark.parametrize(
    "replicate",
    range(20),
)
@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_build_distance_matrix_numpy_triple(
    generations_before_mrca,
    differentia_bit_width,
    estimator,
    prior,
    force_common_ancestry,
    replicate,
    wrap,
):

    common_ancestor = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=differentia_bit_width
    ).CloneNthDescendant(generations_before_mrca)
    population = [
        wrap(common_ancestor.CloneNthDescendant(random.randrange(0, 4))),
        wrap(common_ancestor.CloneNthDescendant(random.randrange(0, 4))),
        wrap(common_ancestor.CloneNthDescendant(random.randrange(0, 4))),
    ]

    d01 = hstrat.estimate_patristic_distance_between(
        population[0], population[1], estimator, prior
    )
    d02 = hstrat.estimate_patristic_distance_between(
        population[0], population[2], estimator, prior
    )
    d12 = hstrat.estimate_patristic_distance_between(
        population[1], population[2], estimator, prior
    )
    assert np.array_equal(
        hstrat.build_distance_matrix_numpy(
            population,
            estimator,
            prior,
            force_common_ancestry=force_common_ancestry,
        ),
        np.array(
            [
                [0.0, d01, d02],
                [d01, 0.0, d12],
                [d02, d12, 0.0],
            ]
        ),
        equal_nan=True,
    )



import numpy as np
import pandas as pd

from ._alifestd_mark_node_depth_asexual import alifestd_mark_node_depth_asexual
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col


def alifestd_unfurl_traversal_postorder_asexual(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
) -> np.ndarray:
    """List `id` values in postorder traversal order.

    The provided dataframe must be asexual.
    """
    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)
    phylogeny_df = alifestd_mark_node_depth_asexual(phylogeny_df, mutate=True)
    postorder_index = np.lexsort(
        (phylogeny_df["ancestor_id"], phylogeny_df["node_depth"]),
    )[::-1]
    id_loc = phylogeny_df.columns.get_loc("id")
    return phylogeny_df.iloc[postorder_index, id_loc].to_numpy()



import pandas as pd

from ._alifestd_mark_oldest_root import alifestd_mark_oldest_root
from ._alifestd_mark_roots import alifestd_mark_roots


def alifestd_join_roots(
    phylogeny_df: pd.DataFrame, mutate: bool = False
) -> pd.DataFrame:
    """Point all other roots to oldest root, measured by lowest `origin_time`
    (if available) or otherwise lowest `id`.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """
    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df = alifestd_mark_roots(phylogeny_df, mutate=True)
    phylogeny_df = alifestd_mark_oldest_root(phylogeny_df, mutate=True)

    if len(phylogeny_df) <= 1:
        return phylogeny_df

    global_root_id = phylogeny_df.loc[
        phylogeny_df["is_oldest_root"].idxmax(), "id"
    ]

    if "ancestor_id" in phylogeny_df:
        phylogeny_df.loc[
            phylogeny_df["is_root"], "ancestor_id"
        ] = global_root_id

    if "ancestor_list" in phylogeny_df:
        phylogeny_df.loc[
            phylogeny_df["is_root"], "ancestor_list"
        ] = f"[{global_root_id}]"

    phylogeny_df["is_root"] = False

    phylogeny_df.loc[
        phylogeny_df["is_oldest_root"], "ancestor_list"
    ] = "[none]"
    phylogeny_df.loc[phylogeny_df["is_oldest_root"], "is_root"] = True

    return phylogeny_df



import typing

from ..._detail import PolicyCouplerBase
from .._PolicySpec import PolicySpec
from .._impl import calc_provided_uncertainty


class CalcRankAtColumnIndex:
    """Functor to provide member function implementation in Policy class."""

    def __init__(
        self: "CalcRankAtColumnIndex",
        policy_spec: typing.Optional[PolicySpec],
    ) -> None:
        pass

    def __eq__(self: "CalcRankAtColumnIndex", other: typing.Any) -> bool:
        return isinstance(other, self.__class__)

    def __call__(
        self: "CalcRankAtColumnIndex",
        policy: PolicyCouplerBase,
        index: int,
        num_strata_deposited: int,
    ) -> int:
        """After n strata have been deposited, what will the rank of the
        stratum at column index k be?

        Enables a HereditaryStratigraphicColumn using this predicate to
        optimize away storage of rank annotations on strata. Takes into the
        account the possibility for in-progress stratum depositions that haven't
        been reflected in num_strata_deposited.
        """
        spec = policy.GetSpec()

        if index == 0:
            # 0th index is always rank 0
            return 0
        elif (
            index
            == policy.CalcNumStrataRetainedExact(num_strata_deposited) - 1
        ):
            # case where index is the very most recent stratum
            return num_strata_deposited - 1
        elif index == policy.CalcNumStrataRetainedExact(num_strata_deposited):
            # in cases where the index is an in-progress
            # deposition rank must be calculated as the rank succeeding the
            # previous stratum's rank
            return (
                self(
                    policy,
                    index - 1,
                    num_strata_deposited,
                )
                + 1
            )
        else:
            # assumes no in-progress stratum depositions that haven't been
            # reflected in num_strata_deposited
            provided_uncertainty = calc_provided_uncertainty(
                spec.GetDepthProportionalResolution(),
                num_strata_deposited,
            )
            return min(index * provided_uncertainty, num_strata_deposited - 1)



import typing

import pandas as pd

from ._alifestd_make_ancestor_list_col import alifestd_make_ancestor_list_col
from ._alifestd_parse_ancestor_ids import alifestd_parse_ancestor_ids


def alifestd_aggregate_phylogenies(
    phylogeny_dfs: typing.List[pd.DataFrame],
    mutate: bool = False,
) -> pd.DataFrame:
    """Concatenate independent phylogenies, reassigning organism ids to
    prevent collisions.

    Inputs dataframe are not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """

    aggregate_least_available_id = 0

    res = []
    for phylogeny_df in phylogeny_dfs:
        if len(phylogeny_df) == 0:
            res.append(phylogeny_df)
            continue

        if not mutate:
            phylogeny_df = phylogeny_df.copy()
        cur_max_id = phylogeny_df["id"].max()

        if aggregate_least_available_id:
            phylogeny_df["id"] += aggregate_least_available_id
            if "ancestor_id" in phylogeny_df:
                phylogeny_df["ancestor_id"] += aggregate_least_available_id
                phylogeny_df[
                    "ancestor_list"
                ] = alifestd_make_ancestor_list_col(
                    phylogeny_df["id"], phylogeny_df["ancestor_id"]
                )
            else:
                phylogeny_df["ancestor_list"] = (
                    phylogeny_df["ancestor_list"]
                    .apply(
                        lambda ancestor_list_str: str(
                            [
                                int(ancestor_id + aggregate_least_available_id)
                                for ancestor_id in alifestd_parse_ancestor_ids(
                                    ancestor_list_str
                                )
                            ]
                        )
                    )
                    .replace("[]", "[none]")
                )

        aggregate_least_available_id += cur_max_id + 1

        res.append(phylogeny_df)

    res_df = pd.concat(res).reset_index(drop=True)

    # if any dataframes didn't provide ancestor_id, nans will occur
    # strip out ancestor_id column in that case
    if "ancestor_id" in res_df and pd.api.types.is_float_dtype(
        res_df["ancestor_id"]
    ):
        res_df.drop("ancestor_id", axis=1, inplace=True)

    return res_df



import typing

import numpy as np
import opytional as opyt

from ...._auxiliary_lib import HereditaryStratigraphicArtifact
from ...estimators import (
    estimate_rank_of_mrca_naive as estimate_rank_of_mrca_naive_,
)
from .._calc_rank_of_mrca_bounds_between import (
    calc_rank_of_mrca_bounds_between,
)


def estimate_rank_of_mrca_naive(
    first: HereditaryStratigraphicArtifact,
    second: HereditaryStratigraphicArtifact,
) -> typing.Optional[float]:
    """Forwards to estimate_rank_of_mrca_naive estimator."""
    # optimization: naive estimator only consideres at the two ranks on either
    # side of the first retained disparity, so only extract those ranks
    rank_of_mrca_bounds = calc_rank_of_mrca_bounds_between(
        first,
        second,
        prior="arbitrary",
        confidence_level=0.49,
    )
    return estimate_rank_of_mrca_naive_(
        # in case of no shared ancestry, rank 0 is first disparity
        opyt.or_value(rank_of_mrca_bounds, [0]),
        p_differentia_collision=np.nan,
        prior="arbitrary",
    )



import typing

import polars as pl


def render_polars_snapshot(
    df: pl.DataFrame,
    df_name: str = "",
    display: typing.Callable = lambda x: x,
) -> typing.Optional[str]:
    """Render an all-columns snapshot of the head of a Polars DataFrame.

    Parameters
    ----------
    df : pl.DataFrame
        The DataFrame to render.
    df_name : str, default ""
        The name of the DataFrame to render.
    display : Callable, optional
        The function to use to display the rendered DataFrame, e.g., print,
        logging.info, etc.

    Returns
    -------
    Optional[str]
        The rendered DataFrame, as a string, if `display` is not used.
    """
    with pl.Config() as cfg:
        cfg.set_tbl_cols(df.lazy().collect_schema().len())
        head = repr(df.lazy().head().collect())
        message = " ".join(
            [
                df_name,
                "df:",
                str(df.lazy().select(pl.len()).collect().item()),
                "rows\n",
                head,
            ]
        )
        return display(message)
